import ast
import copy
import difflib
import os
from collections import Counter

import jellyfish
import jsonlines
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import ttest_ind
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.preprocessing import StandardScaler, minmax_scale

from .loggers import ExperimentLogger
from .misc.ast import analyse_complexity, process_code


def plot_convergence(
    logger: ExperimentLogger,
    metric: str = "Fitness",
    budget: int = 100,
    save: bool = True,
    return_fig: bool = False,
    separate_lines: bool = False,
):
    """
    Plots the convergence of all methods for each problem from an experiment log.

    Args:
        logger (ExperimentLogger): The experiment logger object.
        metric (str, optional): The metric to show as y-axis label.
        save (bool, optional): Whether to save or show the plot.
        return_fig (bool, optional): Whether to return the figure object.
        separate_lines (bool, optional): If True, plots each run using separate line.
    """
    methods, problems = logger.get_methods_problems()

    fig, axes = plt.subplots(
        figsize=(10, 6 * len(problems)), nrows=len(problems), ncols=1
    )
    problem_i = 0
    for problem in problems:
        # Ensure the data is sorted by 'id' and 'fitness'
        data = logger.get_problem_data(problem_name=problem).drop(
            columns=["code"]
        )  # for efficiency we drop code for now
        data.replace([-np.Inf, np.Inf], 0, inplace=True)
        data.fillna(0, inplace=True)

        # Get unique method names
        methods = data["method_name"].unique()
        ax = axes[problem_i] if len(problems) > 1 else axes
        for method in methods:
            method_data = data[data["method_name"] == method].copy()
            method_data = method_data.sort_values(by=["seed", "_id"])

            # Group by 'seed' and calculate the cumulative max fitness
            method_data["cummax_fitness"] = method_data.groupby("seed")[
                "fitness"
            ].cummax()

            # Calculate mean and std deviation of the cumulative max fitness
            if separate_lines:
                for seed in method_data["seed"].unique():
                    seed_data = method_data[method_data["seed"] == seed]
                    ax.plot(
                        seed_data["_id"],
                        seed_data["cummax_fitness"],
                        label=f"{method} (Run {seed})",
                        alpha=0.5,
                    )
            else:
                summary = (
                    method_data.groupby("_id")["cummax_fitness"]
                    .agg(["mean", "std"])
                    .reset_index()
                )
                # Shift X-axis so that _id starts at 1
                summary["_id"] += 1  # Ensures _id starts at 1 instead of 0

                # Plot the mean fitness
                ax.plot(summary["_id"], summary["mean"], label=method)

                # Plot the shaded error region
                ax.fill_between(
                    summary["_id"],
                    summary["mean"] - summary["std"],
                    summary["mean"] + summary["std"],
                    alpha=0.2,
                )

        # Add labels and legend
        ax.set_xlabel("Number of Evaluations")
        if budget is not None:
            ax.set_xlim(1, budget)
        ax.set_ylabel(f"Mean Best {metric}")
        ax.legend(title="Algorithm")
        ax.grid(True)
        ax.set_title(problem)
        problem_i += 1

    plt.tight_layout()
    if save:
        fig.savefig(f"{logger.dirname}/convergence.png")
    elif not return_fig:
        plt.show()
    if return_fig:
        return fig
    plt.close()


def plot_experiment_CEG(
    logger: ExperimentLogger,
    metric: str = "total_token_count",
    budget: int = 100,
    save: bool = True,
    max_seeds=5,
):
    """
    Plot the Code evolution graphs for each run in an experiment, splitted by problem.

    Args:
        logger (ExperimentLogger): The experiment logger object.
        metric (str, optional): The metric to show as y-axis label (should be a statistic from AST / Complexity).
        save (bool, optional): Whether to save or show the plot.
        max_seeds (int, optional): The maximum number of runs to plot.
    """
    methods, problems = logger.get_methods_problems()

    problem_i = 0
    for problem in problems:
        # Ensure the data is sorted by 'id' and 'fitness'
        data = logger.get_problem_data(problem_name=problem)
        data.replace([-np.Inf], 0, inplace=True)
        data.fillna(0, inplace=True)

        # Get unique runs (seeds)
        seeds = data["seed"].unique()
        num_seeds = min(len(seeds), max_seeds)
        # Get unique method names
        methods = data["method_name"].unique()
        fig, axes = plt.subplots(
            figsize=(5 * num_seeds, 5 * len(methods)),
            nrows=len(methods),
            ncols=num_seeds,
            sharey=True,
            squeeze=False,
        )

        method_i = 0
        for method in methods:
            seed_i = 0
            for seed in seeds[:num_seeds]:
                ax = axes[method_i, seed_i]
                run_data = data[
                    (data["method_name"] == method) & (data["seed"] == seed)
                ].copy()
                if len(run_data) == 0:
                    continue
                plot_code_evolution_graphs(
                    run_data,
                    logger.dirname,
                    plot_features=["total_token_count"],
                    save=False,
                    ax=ax,
                )
                ax.set_xlim([0, budget])
                ax.set_xticks(np.arange(0, budget + 1, 10))
                ax.set_xticklabels(np.arange(0, budget + 1, 10))
                ax.set_title(f"{method} run:{seed}")
                if seed_i > 0:
                    ax.set_ylabel(None)
                if method_i < len(methods) - 1:
                    ax.set_xlabel(None)
                seed_i += 1
            method_i += 1

        if save:
            plt.tight_layout()
            fig.savefig(f"{logger.dirname}/CEG_{problem}.png")
        else:
            plt.show()
        plt.close()


def plot_code_evolution_graphs(
    run_data, expfolder=None, plot_features=None, save=True, ax=None
):
    """
    Plots optimization progress and relationships between successive solutions in an
    evolutionary run based on AST metrics. Can plot multiple features or a single feature on a provided axis.

    Args:
        run_data (pandas.DataFrame): DataFrame containing code and fitness values.
        expfolder (str, optional): Folder path where the plots are saved. If None, plots are shown.
        plot_features (list, optional): The features to plot. If None, plots multiple default features.
        save (bool): If True, saves the plots otherwise shows them.
        ax (matplotlib.axes.Axes, optional): The axis to plot on. If None, creates new plots.
    """
    if ax is not None and (plot_features is None or len(plot_features) > 1):
        raise ValueError(
            "If an axis is provided, the length of plot_features must be 1."
        )

    data = run_data.copy()
    data.replace([np.inf, -np.inf], np.nan, inplace=True)
    data["fitness"] = minmax_scale(data["fitness"])
    data.fillna(0, inplace=True)

    complexity_features = [
        "mean_complexity",
        "total_complexity",
        "mean_token_count",
        "total_token_count",
        "mean_parameter_count",
        "total_parameter_count",
    ]

    # Compute AST or complexity-based statistics
    if len(plot_features) == 1 and plot_features[0] in complexity_features:
        analyse_complexity
        df_stats = data["code"].apply(analyse_complexity).apply(pd.Series)
    else:
        df_stats = data["code"].apply(process_code).apply(pd.Series)
    stat_features = df_stats.columns

    # Merge statistics into the dataframe
    data = pd.concat([data, df_stats], axis=1)
    data.fillna(0, inplace=True)

    # Define default features if not provided
    if plot_features is None:
        plot_features = [
            "tsne",
            "pca",
            "total_complexity",
            "total_token_count",
            "total_parameter_count",
        ]
    else:
        plot_features = plot_features

    # Standardize features
    features = data[stat_features].copy()
    scaler = StandardScaler()
    features_scaled = scaler.fit_transform(features)

    # Perform PCA and t-SNE for dimensionality reduction
    pca = PCA(n_components=1)
    pca_projection = pca.fit_transform(features_scaled)
    data["pca"] = pca_projection[:, 0]

    try:
        tsne = TSNE(n_components=1, random_state=42)
        tsne_projection = tsne.fit_transform(features_scaled)
    except Exception:
        # TNSE did not work, probably too small data, just use pca
        tsne_projection = pca_projection

    data["tsne"] = tsne_projection[:, 0]

    # Convert parent IDs from string to list
    data["parent_ids"] = data["parent_ids"].apply(
        lambda x: ast.literal_eval(x) if isinstance(x, str) else x
    )

    # Count occurrences of each parent ID
    parent_counts = Counter(
        parent_id for parent_ids in data["parent_ids"] for parent_id in parent_ids
    )

    data["parent_size"] = data["id"].map(lambda x: parent_counts.get(x, 1) * 2)

    no_axis = False
    if ax is None:
        no_axis = True
    for x_data in plot_features:
        if no_axis:
            fig, ax = plt.subplots(figsize=(8, 5))

        for _, row in data.iterrows():
            for parent_id in row["parent_ids"]:
                if parent_id in data["id"].values:
                    parent_row = data[data["id"] == parent_id].iloc[0]
                    ax.plot(
                        [parent_row["id"], row["id"]],
                        [parent_row[x_data], row[x_data]],
                        "-o",
                        markersize=row["parent_size"],
                        color=plt.cm.viridis(row["fitness"] / max(data["fitness"])),
                    )
                else:
                    ax.plot(
                        row["id"],
                        row[x_data],
                        "o",
                        markersize=row["parent_size"],
                        color=plt.cm.viridis(row["fitness"] / max(data["fitness"])),
                    )

        ax.set_xlabel("Evaluation")
        ax.set_ylabel(x_data.replace("_", " "))
        if no_axis:
            ax.set_ylim(data[x_data].min() - 1, data[x_data].max() + 1)
        ax.set_xticks([])  # Remove x-ticks
        ax.set_xticklabels([])  # Remove x-tick labels
        if no_axis:
            ax.set_title(f"Evolution of {x_data}")
        ax.grid(True)

        if save and expfolder is not None:
            plt.tight_layout()
            plt.savefig(f"{expfolder}/{x_data}_Evolution.png")
        elif ax is None:
            plt.show()
        if ax is None:
            plt.close()


def plot_boxplot_fitness(
    logger: ExperimentLogger, y_label="Fitness", x_label="Method", problems=None
):
    """
    Plots boxplots of fitness grouped by problem_name (subplots) and method_name (categories).
    Each problem has its own subplot, grouped by method_name.

    Args:
        logger
        y_label
        x_label
        problems
    """
    df = logger.get_data().copy()
    # If not already present, create a "fitness" column from the "solution" dictionary
    if "fitness" not in df.columns:
        df["fitness"] = df["solution"].apply(
            lambda sol: sol.get("fitness", float("nan"))
        )

    if problems is None:
        problems = sorted(df["problem_name"].unique())

    # Create subplots, one per problem
    fig, axes = plt.subplots(
        1, len(problems), figsize=(5 * len(problems), 5), sharey=True
    )
    # In case there's only one problem, axes won't be a list
    if len(problems) == 1:
        axes = [axes]

    for i, problem in enumerate(problems):
        subset = df[df["problem_name"] == problem]
        # Plot with Seaborn
        sns.boxplot(x="method_name", y="fitness", data=subset, ax=axes[i])
        axes[i].set_title(problem)
        axes[i].set_xlabel(x_label)
        if i == 0:
            axes[i].set_ylabel(y_label)
        else:
            axes[i].set_ylabel("")
        # Rotate x-axis labels a bit if needed
        axes[i].tick_params(axis="x", rotation=45)

    plt.tight_layout()
    plt.show()


def plot_boxplot_fitness_hue(
    logger: ExperimentLogger,
    y_label="Fitness",
    x_label="Problem",
    hue="method_name",
    x="problem_name",
    problems=None,
):
    """
    Plots boxplots of fitness grouped by `hue` and method_name `x`.
    Produces one plot with grouped boxplots per `x`.
    Args:
        logger
    """
    df = logger.get_data().copy()
    # If not already present, create a "fitness" column from the "solution" dictionary
    if "fitness" not in df.columns:
        df["fitness"] = df["solution"].apply(
            lambda sol: sol.get("fitness", float("nan"))
        )

    if problems is None:
        problems = sorted(df["problem_name"].unique())
    df_filtered = df[df["problem_name"].isin(problems)]

    # Create subplots, one per problem
    fig, axes = plt.subplots(1, 1, figsize=(2.5 * len(problems), 4))

    # Plot with Seaborn
    sns.boxplot(x=x, y="fitness", hue=hue, data=df_filtered, ax=axes)
    axes.set_xlabel(x_label)
    axes.set_ylabel(y_label)

    plt.tight_layout()
    plt.show()


def fitness_table(logger: ExperimentLogger, alpha=0.05, smaller_is_better=False):
    """
    Creates a LaTeX table with rows = methods, columns = problems.
    Each cell shows mean ± std (p=...).
    Cells are bolded if that method is significantly better than all others (p < alpha)
    for the given problem.

    Args:
        logger (ExperimentLogger): the experiment logger to process.
        alpha (float): Significance level for p-value cutoff.
        smaller_is_better (bool): If True, we treat lower fitness as “better.” Otherwise, higher is “better.”
    """
    # Ensure there's a 'fitness' column
    df = logger.get_data().copy()
    if "fitness" not in df.columns:
        df["fitness"] = df["solution"].apply(
            lambda sol: sol.get("fitness", float("nan"))
        )

    # Group data by (problem_name, method_name)
    # We'll store the runs for each combination so we can compute stats and pairwise tests
    grouped = df.groupby(["problem_name", "method_name"])["fitness"]
    stats_dict = {(p, m): grouped.get_group((p, m)).values for p, m in grouped.groups}

    problems = sorted(df["problem_name"].unique())
    methods = sorted(df["method_name"].unique())

    # Prepare a 2D structure for the table: rows=methods, cols=problems
    table_values = []

    for method in methods:
        row_entries = []
        for problem in problems:
            # Retrieve all runs for (problem, method)
            arr = stats_dict.get((problem, method), np.array([]))
            if len(arr) == 0:
                row_entries.append("N/A")
                continue

            mean_val = np.mean(arr)
            std_val = np.std(arr)

            # Compare this method’s distribution to each other method
            # We'll do a t-test and check p-value
            # For "significantly better than all others" we need:
            # 1) The comparison with each other method's distribution has p < alpha
            # 2) The mean is "better" (depending on smaller_is_better).
            all_better = True
            pvals = []
            for other_method in methods:
                if other_method == method:
                    continue
                other_arr = stats_dict.get((problem, other_method), np.array([]))
                if len(other_arr) == 0:
                    # If there’s no data for the other method, skip
                    continue

                # Mean comparison
                other_mean = np.mean(other_arr)
                is_better = (
                    (mean_val < other_mean)
                    if smaller_is_better
                    else (mean_val > other_mean)
                )
                if not is_better:
                    all_better = False

                # Significance test
                # (We could use ttest_ind, Mann-Whitney U, etc. Here we do ttest_ind.)
                _, pval = ttest_ind(arr, other_arr, equal_var=False)
                pvals.append(pval)

            # We'll store the *maximum* p-value among all pairwise comparisons,
            # because for the method to be "significantly better than all others",
            # *every* p-value must be below alpha.
            max_p = max(pvals) if pvals else 1.0

            cell_str = f"{mean_val:.2f} ± {std_val:.2f} (p={max_p:.3f})"
            if all_better and (max_p < alpha):
                cell_str = "\\textbf{" + cell_str + "}"
            row_entries.append(cell_str)
        table_values.append(row_entries)

    # Create a DataFrame of the final strings so we can export to LaTeX nicely
    table_df = pd.DataFrame(table_values, index=methods, columns=problems)

    return table_df


def plot_token_usage(
    logger: ExperimentLogger, save: bool = True, return_fig: bool = False
):
    """Plot total tokens used per method/problem from an experiment logger."""

    df = logger.get_data().copy()
    token_records = []
    for _, row in df.iterrows():
        tokens = 0
        if "log_dir" in row:
            convo_path = os.path.join(
                logger.dirname, row["log_dir"], "conversationlog.jsonl"
            )
            if os.path.isfile(convo_path):
                with jsonlines.open(convo_path) as f:
                    for line in f:
                        tokens += line.get("tokens", 0)
        token_records.append(
            {
                "method_name": row["method_name"],
                "problem_name": row["problem_name"],
                "tokens": tokens,
            }
        )
    token_df = pd.DataFrame(token_records)
    summary = (
        token_df.groupby(["problem_name", "method_name"])["tokens"].sum().reset_index()
    )

    fig, ax = plt.subplots(figsize=(7, 5))
    sns.barplot(
        data=summary,
        x="problem_name",
        y="tokens",
        hue="method_name",
        ax=ax,
    )
    ax.set_xlabel("Problem")
    ax.set_ylabel("Total Tokens")
    plt.tight_layout()

    if save:
        fig.savefig(os.path.join(logger.dirname, "token_usage.png"))
    elif not return_fig:
        plt.show()
    if return_fig:
        return fig
    plt.close()



from iohblade.experiment import Experiment
from iohblade.llm import Gemini_LLM, Ollama_LLM, OpenAI_LLM
from iohblade.methods import LLaMEA, RandomSearch
from iohblade.loggers import ExperimentLogger
from iohblade.problems import Kerneltuner
import numpy as np
import os

if __name__ == "__main__": # prevents weird restarting behaviour
    api_key = os.getenv("OPENAI_API_KEY")
    api_key_gemini = os.getenv("GEMINI_API_KEY")

    llm1 = Gemini_LLM(api_key_gemini, "gemini-2.0-flash")
    llm2 = OpenAI_LLM(api_key,"o3-2025-04-16", temperature=1.0)
    llm5 = OpenAI_LLM(api_key,"o4-mini-2025-04-16", temperature=1.0) #Done
    budget = 100

    mutation_prompts = [
        "Refine the strategy of the selected solution to improve it.",  # small mutation
        "Generate a new algorithm that is different from the algorithms you have tried before.", #new random solution
        "Refine and simplify the selected algorithm to improve it.", #simplify
    ]

    for llm in [llm5]: #, llm1
        LLaMEA_method = LLaMEA(llm, budget=budget, name="o4 (no info)", mutation_prompts=mutation_prompts, n_parents=4, n_offspring=12, elitism=True)

        methods = [LLaMEA_method]
        logger = ExperimentLogger("results/kerneltuner-o4-no-info")
        problems = [
            Kerneltuner(
                gpus=["A100", "A4000", "MI250X"], #, "A4000", "MI250X"
                kernels=["gemm"],
                name="kerneltuner-gemm",
                eval_timeout=600, #5 minutes
                budget=1000,
                cache_dir="/data/neocortex/repos/benchmark_hub/",
                extra_info=False),
            Kerneltuner(
                gpus=["A100", "A4000", "MI250X"],
                kernels=["convolution"],
                name="kerneltuner-convolution",
                eval_timeout=600,
                budget=1000,
                cache_dir="/data/neocortex/repos/benchmark_hub/",
                extra_info=False),
            Kerneltuner(
                gpus=["A100", "A4000", "MI250X"],
                kernels=["dedispersion"],
                name="kerneltuner-dedispersion",
                eval_timeout=600,
                budget=1000,
                cache_dir="/data/neocortex/repos/benchmark_hub/",
                extra_info=False),
            Kerneltuner(
                gpus=["A100", "A4000", "MI250X"],
                kernels=["hotspot"],
                name="kerneltuner-hotspot",
                eval_timeout=600,
                budget=1000,
                cache_dir="/data/neocortex/repos/benchmark_hub/",
                extra_info=False),
            
            ]
        experiment = Experiment(methods=methods, problems=problems, runs=3, budget = budget, show_stdout=True, exp_logger=logger) #normal run
        experiment() #run the experiment


import numpy as np
from modcma import c_maes


class ModularCMAES:
    """
    Baseline implementation of CMA-ES with active update.
    Can be extended later to cover all Modular CMAES options.
    """

    def __init__(self, budget=10000, dim=10, **kwargs):
        self.budget = budget
        self.dim = dim
        # Instantate a modules object
        self.modules = c_maes.parameters.Modules()
        self.modules.matrix_adaptation = c_maes.options.MatrixAdaptationType.COVARIANCE
        self.modules.active = True
        # Create a settings object, here also optional parameters such as sigma0 can be specified
        self.settings = c_maes.parameters.Settings(dim, self.modules, **kwargs)
        # Create a parameters object
        self.parameters = c_maes.Parameters(self.settings)
        # Pass the parameters object to the ModularCMAES optimizer class
        self.cma = c_maes.ModularCMAES(self.parameters)

    def __call__(self, func):
        return self.cma.run(func)



import json
import os

# Use a non-interactive backend for CI
import matplotlib
import numpy as np
import pandas as pd
import pytest

matplotlib.use("Agg")
from unittest.mock import MagicMock

import matplotlib.pyplot as plt

# Adjust imports to match your actual package structure
from iohblade import (
    fitness_table,
    plot_boxplot_fitness,
    plot_boxplot_fitness_hue,
    plot_code_evolution_graphs,
    plot_convergence,
    plot_experiment_CEG,
    plot_token_usage,
)

# ------------------------------------------------------------------------
# Mock objects and data
# ------------------------------------------------------------------------


class MockExperimentLogger:
    """A minimal mock that simulates the interface needed by your plots."""

    def __init__(self):
        # Some dummy data with the columns your plot functions expect
        self.mock_data = pd.DataFrame(
            {
                "method_name": ["m1", "m1", "m2", "m2"],
                "problem_name": ["p1", "p1", "p1", "p1"],
                "seed": [0, 0, 1, 1],
                "_id": [0, 1, 2, 3],
                "id": ["0", "1", "2", "3"],
                "fitness": [0.0, 1.5, 0.5, 2.0],
                "code": [
                    "print('hello')",
                    "def f(): pass",
                    "def g(): pass",
                    """
def run(budget=100, dim=5):
    for i in range(budget):
        print(i)
    return i+1
""",
                ],
                "parent_ids": ["[]", '["0"]', '["1","0"]', '["2"]'],
                "log_dir": [f"run{i}" for i in range(4)],
            }
        )
        # Additional problems, if you want to test multi-problem plotting
        self.mock_data2 = pd.DataFrame(
            {
                "method_name": ["m1", "m2", "m1", "m2"],
                "problem_name": ["p2", "p2", "p2", "p2"],
                "seed": [0, 0, 1, 1],
                "_id": [0, 1, 2, 3],
                "id": ["0", "1", "2", "3"],
                "fitness": [1.0, 2.0, 3.0, 4.0],
                "code": ["code e", "code f", "code g", "code h"],
                "parent_ids": ["[]", "[]", "[]", "[]"],
                "log_dir": [f"run{i+4}" for i in range(4)],
            }
        )

    @property
    def dirname(self):
        return "mock_dir"

    def get_methods_problems(self):
        """Return the unique methods and problems in the log."""
        methods = self.mock_data["method_name"].unique().tolist()
        problems = list(
            set(self.mock_data["problem_name"].unique())
            | set(self.mock_data2["problem_name"].unique())
        )
        return (methods, sorted(problems))

    def get_problem_data(self, problem_name):
        """Return a dataframe filtered by problem name."""
        # Combine both mock_data sets so that each problem can be tested
        combined = pd.concat([self.mock_data, self.mock_data2], ignore_index=True)
        return combined[combined["problem_name"] == problem_name].copy()

    def get_data(self):
        """Return a combined dataframe to mimic the real logger's get_data()."""
        return pd.concat([self.mock_data, self.mock_data2], ignore_index=True)


@pytest.fixture
def mock_logger(tmp_path):
    # If the plotting code tries to save a file, it’ll do so in this tmp path
    os.chdir(tmp_path)  # switch working dir to temp
    logger = MockExperimentLogger()
    os.makedirs(logger.dirname, exist_ok=True)
    df = pd.concat([logger.mock_data, logger.mock_data2], ignore_index=True)
    for ld in df["log_dir"]:
        run_dir = os.path.join(logger.dirname, ld)
        os.makedirs(run_dir, exist_ok=True)
        with open(os.path.join(run_dir, "conversationlog.jsonl"), "w") as f:
            f.write(json.dumps({"tokens": 5, "content": "x"}) + "\n")
    return logger


# ------------------------------------------------------------------------
# Tests for each plotting function
# ------------------------------------------------------------------------
def test_plot_convergence(mock_logger):
    # Just ensure it runs without exception and creates a figure
    plot_convergence(logger=mock_logger, metric="Fitness", budget=3, save=False)
    # We can grab the figure with plt.gcf() if needed
    fig = plt.gcf()
    assert isinstance(fig, plt.Figure)
    plt.close(fig)


def test_plot_experiment_CEG(mock_logger):
    # Ensure the function runs. It creates multiple subplots for multiple problems.
    plot_experiment_CEG(
        logger=mock_logger,
        metric="total_token_count",
        budget=3,
        save=False,
        max_seeds=2,
    )
    fig = plt.gcf()
    assert isinstance(fig, plt.Figure)
    plt.close(fig)


def test_plot_code_evolution_graphs_single_feature():
    # Minimal run_data
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "parent_ids": ["[]", "[0]", "[1]"],
            "fitness": [1.0, 2.0, 3.0],
            "code": [
                """
def runa(budget=2, dim=5):
    for i in range(budget):
        print(i)
        print(i+1)
    return i+3
""",
                """
def runb(budget=100, dim=5):
    for i in range(budget):
        print(i)
    return i+1
""",
                """
def runc(budget=100, dim=5):
    for j in range(dim):
        print(j)
    return j
""",
            ],
        }
    )
    # We pass a single feature
    fig, ax = plt.subplots()
    plot_code_evolution_graphs(
        run_data=df,
        expfolder=None,
        plot_features=["total_token_count"],  # single feature
        save=False,
        ax=ax,
    )
    # If the function runs, we have a figure
    assert isinstance(plt.gcf(), plt.Figure)
    plt.close(fig)


def test_plot_code_evolution_graphs_multiple_features():
    # If we pass multiple features, we must not pass an external ax
    df = pd.DataFrame(
        {
            "id": ["0", "1", "2"],
            "parent_ids": ["[]", '["0"]', '["1"]'],
            "fitness": [1.0, 2.0, 3.0],
            "code": ["print('hello')", "def f(): pass", "def g(): print('C')"],
        }
    )
    plot_code_evolution_graphs(
        run_data=df,
        expfolder=None,
        plot_features=["pca", "tsne"],  # multiple features
        save=False,
        ax=None,
    )
    assert isinstance(plt.gcf(), plt.Figure)
    plt.close("all")


def test_plot_boxplot_fitness(mock_logger):
    # The code references the "fitness" column, so we just run it:
    plot_boxplot_fitness(
        logger=mock_logger, y_label="Fitness", x_label="Method", problems=["p1", "p2"]
    )
    assert isinstance(plt.gcf(), plt.Figure)
    plt.close("all")


def test_plot_boxplot_fitness_hue(mock_logger):
    plot_boxplot_fitness_hue(
        logger=mock_logger, x_label="Problem", hue="method_name", x="problem_name"
    )
    assert isinstance(plt.gcf(), plt.Figure)
    plt.close("all")


def test_fitness_table(mock_logger):
    # Check that it returns a DataFrame with methods x problems
    table_df = fitness_table(mock_logger, alpha=0.05, smaller_is_better=False)
    assert isinstance(table_df, pd.DataFrame)
    methods, problems = mock_logger.get_methods_problems()
    # The shape should be (#methods) x (#problems) if everything's consistent
    assert table_df.shape == (len(methods), len(problems))
    # Optionally check for bold text in some cells, etc.
    # e.g., to confirm the string format:
    for val in table_df.values.flatten():
        assert isinstance(val, str)
    # No figure is created for a table, so nothing to close.


def test_plot_token_usage(mock_logger):
    plot_token_usage(mock_logger, save=False)
    assert isinstance(plt.gcf(), plt.Figure)
    plt.close("all")



import numpy as np
import PyMoosh as pm

from .photonic_problem import photonic_problem


class ellipsometry(photonic_problem):
    def __init__(
        self,
        mat_env,
        mat_substrate,
        nb_layers,
        min_thick,
        max_thick,
        min_eps,
        max_eps,
        wavelengths,
        angle,
    ):
        self.n = nb_layers * 2
        self.mat_env = mat_env
        self.mat_substrate = mat_substrate
        self.nb_layers = nb_layers
        self.min_thick = min_thick
        self.max_thick = max_thick
        self.min_eps = min_eps
        self.max_eps = max_eps
        self.wavelengths = wavelengths
        self.angle = angle
        self.lb = np.array(
            [self.min_eps] * self.nb_layers + [self.min_thick] * self.nb_layers
        )
        self.ub = np.array(
            [self.max_eps] * self.nb_layers + [self.max_thick] * self.nb_layers
        )
        x_ref = np.random.uniform(self.lb, self.ub, self.n)
        struct_ref = self.setup_structure(x_ref)
        self.ref_ellipso = np.zeros(len(self.wavelengths), dtype=complex)
        for i, wav in enumerate(self.wavelengths):
            r_s, _, _, _ = pm.coefficient(struct_ref, wav, self.angle, 0)
            r_p, _, _, _ = pm.coefficient(struct_ref, wav, self.angle, 1)
            self.ref_ellipso[i] = r_p / r_s

    def setup_structure(self, x):
        """helper to create pymoosh structure object with user-defined
        thicknesses and materials

        Args:
            X (list): long list of material permittivities and thicknesses
                (first half / second half)
            mat_env (float, str): material of environment (above stack)
            mat_substrate (float, str): material of substrate (below stack)

        Returns:
            PyMoosh.structure: multi-layer structure object
        """
        x = list(x)
        # available materials
        materials = (
            [self.mat_env] + [_m for _m in x[: self.nb_layers]] + [self.mat_substrate]
        )
        # material sequence of layer-stack
        stack = [i for i in range(self.nb_layers + 2)]
        # thicknesses of layers
        thicknesses = np.array([0] + [_t for _t in x[self.nb_layers :]] + [0])
        structure = pm.Structure(materials, stack, np.array(thicknesses), verbose=False)
        return structure

    def __call__(self, x):
        """cost function: MAE between simulated and measured (ref) spectrum

        Args:
            x (list): materials (first half) & thicknesses (second half) of all
                layers
            ref_ellipso (np.ndarray): reference spectrum at `eval_wls`
            wavelengths (np.ndarray): wavelengths to evaluate

        Returns:
            float: Reflectivity at target wavelength
        """
        x = np.clip(x, self.lb, self.ub)
        structure = self.setup_structure(x)
        # the actual PyMoosh reflectivity simulation
        ellips = np.zeros(len(self.wavelengths), dtype=complex)
        for i, wav in enumerate(self.wavelengths):
            r_s, _, _, _ = pm.coefficient(structure, wav, self.angle, 0)
            r_p, _, _, _ = pm.coefficient(structure, wav, self.angle, 1)
            ellips[i] = r_p / r_s
        # diff = ellips - ref_ellipso
        cost = np.sum(np.abs(ellips - self.ref_ellipso))
        return cost



import os
import shutil
from unittest.mock import MagicMock

import pytest

from iohblade import LLM, Method, Problem
from iohblade.experiment import Experiment, MA_BBOB_Experiment
from iohblade.loggers import ExperimentLogger


@pytest.fixture
def cleanup_tmp_dir():
    # Creates a temporary directory for tests, yields its name, then cleans up
    dirname = "test_results"
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    yield dirname
    # Cleanup
    if os.path.exists(dirname):
        shutil.rmtree(dirname)


def test_ma_bbob_experiment_init(cleanup_tmp_dir):
    class DummyMethod(Method):
        def __call__(self, problem):
            pass

        def to_dict(self):
            return {}

    class DummyLLM(LLM):
        def _query(self, s):
            return "res"

    llm = DummyLLM(api_key="", model="")
    methods = [DummyMethod(llm, 10, name="m1")]

    exp = MA_BBOB_Experiment(
        methods,
        runs=2,
        budget=50,
        dims=[2, 3],
        budget_factor=1000,
        exp_logger=ExperimentLogger(os.path.join(cleanup_tmp_dir, "mabbob_experiment")),
    )
    assert len(exp.problems) == 1  # Just one MA_BBOB instance
    assert exp.runs == 2
    assert exp.budget == 50


def test_experiment_run(cleanup_tmp_dir):
    class DummyExp(Experiment):
        def __call__(self):
            self.exp_logger.add_run(
                self.methods[0],
                self.problems[0],
                self.methods[0].llm,
                MagicMock(),
                log_dir="test",
                seed=0,
            )

    class DummyMethod(Method):
        def __call__(self, problem):
            pass

        def to_dict(self):
            return {}

    class DummyProblem(Problem):
        def get_prompt(self):
            return "Problem prompt"

        def evaluate(self, s):
            return s

        def test(self, s):
            return s

        def to_dict(self):
            return {}

    class DummyLLM(LLM):
        def _query(self, session_messages):
            return "response"

    l = DummyLLM("", "")
    m = DummyMethod(l, 5, name="DMethod")
    p = DummyProblem()

    exp = DummyExp(
        methods=[m],
        problems=[p],
        exp_logger=ExperimentLogger(os.path.join(cleanup_tmp_dir, "mabbob_experiment")),
    )
    exp()  # call
    # Check something about the exp_logger, or just ensure it doesn't crash



import os
import shutil

import pytest

from iohblade import Method, Problem, Solution
from iohblade.loggers import ExperimentLogger, RunLogger


@pytest.fixture
def cleanup_tmp_dir():
    # Creates a temporary directory for tests, yields its name, then cleans up
    dirname = "test_results"
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    yield dirname
    # Cleanup
    if os.path.exists(dirname):
        shutil.rmtree(dirname)


def test_experiment_logger_add_run(cleanup_tmp_dir):
    exp_logger = ExperimentLogger(name=os.path.join(cleanup_tmp_dir, "my_experiment"))
    from iohblade.llm import LLM
    from iohblade.method import Method
    from iohblade.problem import Problem

    class DummyMethod(Method):
        def __call__(self, problem):
            pass

        def to_dict(self):
            return {}

    class DummyProblem(Problem):
        def get_prompt(self):
            return "prompt"

        def evaluate(self, s):
            return s

        def test(self, s):
            return s

        def to_dict(self):
            return {}

    class DummyLLM(LLM):
        def _query(self, s):
            return "res"

    method = DummyMethod(None, 100, name="dummy_method")
    problem = DummyProblem()
    llm = DummyLLM("", "")
    sol = Solution()
    exp_logger.add_run(method, problem, llm, sol, log_dir="dummy_dir", seed=42)

    # Check the log file
    log_file = os.path.join(exp_logger.dirname, "experimentlog.jsonl")
    with open(log_file, "r") as f:
        contents = f.read()
    assert "dummy_method" in contents
    expected_rel = os.path.relpath("dummy_dir", exp_logger.dirname)
    assert expected_rel in contents
    assert '"seed": 42' in contents


def test_run_logger_log_individual(cleanup_tmp_dir):
    run_logger = RunLogger(name="test_run", root_dir=cleanup_tmp_dir)
    sol = Solution(name="test_solution")
    run_logger.log_individual(sol)
    # Check existence of log.jsonl
    log_file = os.path.join(run_logger.dirname, "log.jsonl")
    assert os.path.exists(log_file)
    with open(log_file, "r") as f:
        contents = f.read()
    assert "test_solution" in contents


def test_run_logger_budget_exhausted(cleanup_tmp_dir):
    run_logger = RunLogger(name="test_run", root_dir=cleanup_tmp_dir, budget=1)
    sol = Solution(name="test_solution")
    run_logger.log_individual(sol)
    assert run_logger.budget_exhausted() is True


def test_experiment_logger_get_data(cleanup_tmp_dir):
    exp_logger = ExperimentLogger(
        name=os.path.join(cleanup_tmp_dir, "my_experiment_data")
    )
    # Write a dummy JSON line
    log_file = os.path.join(exp_logger.dirname, "experimentlog.jsonl")
    with open(log_file, "w") as f:
        f.write('{"method_name":"methodA","problem_name":"problemX"}\n')
        f.write('{"method_name":"methodB","problem_name":"problemY"}\n')
    df = exp_logger.get_data()
    assert len(df) == 2
    assert "method_name" in df.columns
    assert "problem_name" in df.columns


def test_read_multiple_log_dirs(cleanup_tmp_dir):
    exp_logger1 = ExperimentLogger(name=os.path.join(cleanup_tmp_dir, "multi_dir1"))
    exp_logger2 = ExperimentLogger(name=os.path.join(cleanup_tmp_dir, "multi_dir2"))

    # create a sample log file in each directory
    run_logger = RunLogger(name="test_run1", root_dir=exp_logger1.dirname)
    sol = Solution(name="test_solution").set_scores(1.0, feedback="Test feedback")
    run_logger.log_individual(sol)
    # Check existence of log.jsonl
    log_file1 = os.path.join(run_logger.get_log_dir(), "log.jsonl")
    assert os.path.exists(log_file1)

    run_logger2 = RunLogger(name="test_run2", root_dir=exp_logger2.dirname)
    sol = Solution(name="test_solution").set_scores(1.0, feedback="Test feedback")
    run_logger2.log_individual(sol)
    # Check existence of log.jsonl
    log_file2 = os.path.join(run_logger2.get_log_dir(), "log.jsonl")
    assert os.path.exists(log_file2)

    log_file = os.path.join(exp_logger1.dirname, "experimentlog.jsonl")
    with open(log_file, "w") as f:
        f.write(
            f'{{"method_name":"methodA","problem_name":"problemX", "log_dir":"run-test_run1", "seed":"0"}}\n'
        )
        f.write(
            f'{{"method_name":"methodB","problem_name":"problemY", "log_dir":"run-test_run1", "seed":"0"}}\n'
        )

    log_file = os.path.join(exp_logger2.dirname, "experimentlog.jsonl")
    with open(log_file, "w") as f:
        f.write(
            f'{{"method_name":"methodC","problem_name":"problemX", "log_dir":"run-test_run2", "seed":"1"}}\n'
        )
        f.write(
            f'{{"method_name":"methodD","problem_name":"problemY", "log_dir":"run-test_run2", "seed":"1"}}\n'
        )

    # now create a new ExperimentLogger that reads both directories.
    explogger_read = ExperimentLogger(name=exp_logger1.dirname, read=True)
    explogger_read.add_read_dir(exp_logger2.dirname)
    data = explogger_read.get_data()
    assert len(data) == 4
    problem_data = explogger_read.get_problem_data("problemX")
    assert len(problem_data) == 2, "Expected 2 entries for problemX, got {}".format(
        len(problem_data)
    )


def test_experiment_logger_get_methods_problems(cleanup_tmp_dir):
    """`get_methods_problems` should return the unique sets of methods and problems
    across all read-in experiment directories, ignoring dirs with no log."""
    import os

    from iohblade.loggers import ExperimentLogger

    # Make three experiment dirs: two with logs, one empty
    dir1 = os.path.join(cleanup_tmp_dir, "exp1")
    dir2 = os.path.join(cleanup_tmp_dir, "exp2")

    os.makedirs(dir1, exist_ok=True)
    os.makedirs(dir2, exist_ok=True)

    # Write JSON-lines logs
    with open(os.path.join(dir1, "experimentlog.jsonl"), "w") as f:
        f.write('{"method_name":"methodA","problem_name":"problemX"}\n')
        f.write('{"method_name":"methodB","problem_name":"problemY"}\n')
    with open(os.path.join(dir2, "experimentlog.jsonl"), "w") as f:
        f.write('{"method_name":"methodC","problem_name":"problemX"}\n')
        # Duplicate methodA, new problemZ – should be deduped
        f.write('{"method_name":"methodA","problem_name":"problemZ"}\n')

    # Read the dirs
    exp_logger = ExperimentLogger(name=dir1, read=True)
    exp_logger.add_read_dir(dir2)

    methods, problems = exp_logger.get_methods_problems()

    assert set(methods) == {"methodA", "methodB", "methodC"}
    assert set(problems) == {"problemX", "problemY", "problemZ"}


def test_start_progress_and_restart(tmp_path):
    class DummyMethod(Method):
        def __call__(self, problem):
            pass

        def to_dict(self):
            return {}

    class DummyProblem(Problem):
        def get_prompt(self):
            return "prompt"

        def evaluate(self, s):
            return s

        def test(self, s):
            return s

        def to_dict(self):
            return {}

    m = DummyMethod(None, 1, name="m")
    p = DummyProblem(name="p")
    logger_dir = tmp_path / "exp"
    logger = ExperimentLogger(name=str(logger_dir))
    logger.start_progress(1, methods=[m], problems=[p], seeds=[0], budget=1)

    # Restart with same config should succeed
    logger2 = ExperimentLogger(name=str(logger_dir))
    logger2.start_progress(1, methods=[m], problems=[p], seeds=[0], budget=1)
    assert logger2.is_run_pending(m, p, 0)


def test_start_progress_mismatch(tmp_path):
    class DummyMethod(Method):
        def __call__(self, problem):
            pass

        def to_dict(self):
            return {}

    class DummyProblem(Problem):
        def get_prompt(self):
            return "prompt"

        def evaluate(self, s):
            return s

        def test(self, s):
            return s

        def to_dict(self):
            return {}

    m1 = DummyMethod(None, 1, name="m1")
    m2 = DummyMethod(None, 1, name="m2")
    p = DummyProblem(name="p")
    logger_dir = tmp_path / "exp"
    logger = ExperimentLogger(name=str(logger_dir))
    logger.start_progress(1, methods=[m1], problems=[p], seeds=[0], budget=1)

    logger2 = ExperimentLogger(name=str(logger_dir))
    with pytest.raises(ValueError):
        logger2.start_progress(1, methods=[m2], problems=[p], seeds=[0], budget=1)


def test_open_run_create_and_restart(tmp_path):
    """Check that `open_run` correctly creates, restarts, and updates progress."""

    # --- dummy objects -------------------------------------------------
    class DummyMethod(Method):
        def __call__(self, problem):
            pass

        def to_dict(self):
            return {}

    class DummyProblem(Problem):
        def get_prompt(self):
            return "prompt"

        def evaluate(self, s):  # pragma: no cover
            return s

        def test(self, s):  # pragma: no cover
            return s

        def to_dict(self):
            return {}

        # Keep a reference so we can assert the logger was set.
        def set_logger(self, logger):
            self._logger = logger
            super().set_logger(logger)

    # --- setup ---------------------------------------------------------
    exp_dir = tmp_path / "exp"
    logger = ExperimentLogger(name=str(exp_dir))

    m = DummyMethod(None, 1, name="methodX")
    p = DummyProblem(name="problemY")

    # ------------------------------------------------------------------
    # 1. First call should create a run directory and a progress entry
    # ------------------------------------------------------------------
    run1 = logger.open_run(m, p, budget=5, seed=0)
    entry = logger._get_run_entry("methodX", "problemY", 0)

    assert entry is not None, "Progress entry not created"
    assert entry["evaluations"] == 0
    assert entry["end_time"] is None
    assert os.path.isdir(run1.dirname)
    assert entry["log_dir"] == os.path.relpath(run1.dirname, logger.dirname)
    # Logger should be attached to the problem
    assert getattr(p, "_logger", None) is run1

    # The callback should increment `evaluations`
    run1.log_individual(Solution(name="sol"))
    assert entry["evaluations"] == 1

    first_dir = run1.dirname
    assert os.path.exists(first_dir)

    # ------------------------------------------------------------------
    # 2. Second call with same identifiers (unfinished run) should
    #    delete old dir, reset evaluations, and create a new dir
    # ------------------------------------------------------------------
    run2 = logger.open_run(m, p, budget=5, seed=0)
    entry_after = logger._get_run_entry("methodX", "problemY", 0)

    assert entry_after["evaluations"] == 0, "Evaluations were not reset"
    assert entry_after["log_dir"] == os.path.relpath(run2.dirname, logger.dirname)
    assert getattr(p, "_logger", None) is run2

    # ------------------------------------------------------------------
    # 3. Different seed should create an independent progress entry
    # ------------------------------------------------------------------
    run3 = logger.open_run(m, p, budget=5, seed=1)
    new_entry = logger._get_run_entry("methodX", "problemY", 1)

    assert new_entry is not None
    assert new_entry is not entry_after  # distinct entry
    assert os.path.isdir(run3.dirname)



import iohinspector
import polars as pl
from iohblade.behaviour_metrics import compute_behavior_metrics
from tqdm import tqdm
import pandas as pd
import numpy as np
from iohblade.loggers import ExperimentLogger
from iohblade import plot_convergence, plot_experiment_CEG, plot_boxplot_fitness_hue, plot_boxplot_fitness, fitness_table
import os

data_dir = "/data/neocortex/BBOB-2"

logger = ExperimentLogger(data_dir, True)

data = logger.get_data()


tqdm.pandas(desc="Processing runs")

func_ids        = [1, 3, 6, 8, 10, 13, 15, 17, 21, 23]
runs_per_func   = 5          # we know each function has 5 runs

def avg_auc_for_fid(aucs: list[float], fid: int, runs_per_func=5, func_ids=[1, 3, 6, 8, 10, 13, 15, 17, 21, 23]) -> float:
    """
    aucs = 50-long list in metadata.
    Take the five entries belonging to the requested fid and average them.
    """
    if len(aucs) < 50:
        return 0
    block = func_ids.index(fid) * runs_per_func
    slice_ = aucs[block : block + runs_per_func]
    return float(np.mean(slice_))

def process_run(row, func_ids= [1, 3, 6, 8, 10, 13, 15, 17, 21, 23], runs_per_func=5, root=f"{data_dir}/ioh/"):
    rows = []
    # each algorithm has 50 different runs = 50 different directories we need to process
    algid = row['id']
    fitness = row['fitness']
    aucs = row['aucs_list']
    method_name = row['method_name']
    problem_name = row['problem_name']

    counter = 0
    for fid in func_ids:
        for run in range(runs_per_func):
            
            path = f"{root}{algid}"
            if counter > 0:
                path = f"{root}{algid}-{counter}"
            # each algorithm has 50 different runs = 50 different directories we need to process
            if not os.path.exists(path):
                continue
            manager = iohinspector.DataManager()
            manager.add_folder(path)
            # load run data for one folder
            df = manager.load(monotonic=False, include_meta_data=True)
            metrics = compute_behavior_metrics(df)
            counter += 1

            f_fid = avg_auc_for_fid(aucs, fid, runs_per_func=runs_per_func, func_ids=func_ids)
            metrics.update(
                {
                    "id":           algid,
                    "fid":          fid,
                    "fitness_fid":  f_fid,
                    "method_name":  method_name,
                    "problem_name": problem_name,
                    "seed" :        row['seed'], 
                    "_id" :         row['_id'],
                }
            )
            rows.append(metrics)
    return pd.DataFrame(rows)

def get_aucs(d):
    """Return the list under key 'aucs' (or [] if anything is wrong)."""
    try:
        return d.get("aucs", [])
    except (ValueError, SyntaxError):
        return []


methods, problems = logger.get_methods_problems()
print(methods)
print(problems)

for problem in problems:
    print(f"Problem: {problem}")
    df = logger.get_problem_data(problem_name=problem)
    print(df.columns)

    df["aucs_list"] = df["metadata"].apply(get_aucs)


    frames = []

    for _, algo_row in tqdm(df.iterrows(),  # <‑‑ your dataframe of algorithms
                            total=len(df),
                            desc="algorithms"):
        try:
            df_runs = process_run(algo_row)      # your function above
            frames.append(df_runs)
        except Exception as e:
            print(f"⚠︎  {algo_row['id']} skipped ({e})")

    all_runs = pd.concat(frames, ignore_index=True)

    print(f"✅  built dataframe with {len(all_runs)} runs")

    print(all_runs.columns)
    all_runs.to_pickle(f"./{problem}2.pkl")


import json
import os
from datetime import datetime

import mlflow
import mlflow.exceptions
import mlflow.pyfunc
import pandas as pd
from ConfigSpace.read_and_write import json as cs_json

from ..llm import LLM
from ..method import Method
from ..problem import Problem
from ..solution import Solution
from ..utils import convert_to_serializable
from .base import ExperimentLogger, RunLogger


class MLFlowExperimentLogger(ExperimentLogger):
    """
    An ExperimentLogger subclass that keeps the original file-based logging,
    and also logs runs to MLflow. The idea is:
      - Call open_run() at the start of a run (this calls mlflow.start_run()).
      - Run your optimization, logging step-level data via MLFlowRunLogger.
      - Call add_run() at the end, which logs final info and ends the MLflow run.
    """

    def __init__(self, name="", read=False, mlflow_tracking_uri=None):
        """
        Args:
            name (str): The name of the experiment (used as the MLflow experiment name).
            read (bool): If True, read the existing log directory for file-based logs only.
            mlflow_tracking_uri (str): The MLflow Tracking URI (e.g. 'file:/tmp/mlruns',
                                       or your remote server).
        """
        super().__init__(name=name, read=read)
        # If you want to store the logs in some custom place
        if mlflow_tracking_uri:
            mlflow.set_tracking_uri(mlflow_tracking_uri)

        # Create or retrieve a named MLflow experiment
        try:
            self.experiment_id = mlflow.create_experiment(name)
        except mlflow.exceptions.MlflowException:
            # Already exists
            self.experiment_id = mlflow.get_experiment_by_name(name).experiment_id

        self._mlflow_run_active = False  # Track if we have an active run
        self._current_run_id = None

    def open_run(self, method, problem, budget=100, seed=0):
        """
        Opens (starts) a new MLflow run for logging.
        Typically call this right before your run, so that the RunLogger can log step data.
        """
        if self._mlflow_run_active:
            print("Warning: An MLflow run is already active. Not starting a new one.")
            return
        run_name = f"{method.name}-{problem.name}-{seed}"

        run = mlflow.start_run(experiment_id=self.experiment_id, run_name=run_name)
        self._mlflow_run_active = True
        self._current_run_id = run.info.run_id

        self.run_logger = MLFlowRunLogger(
            name=run_name,
            root_dir=self.dirname,
            budget=budget,
            progress_callback=lambda: self.increment_eval(
                method.name, problem.name, seed
            ),
        )
        problem.set_logger(self.run_logger)
        return self.run_logger

    def add_run(
        self,
        method: Method,
        problem: Problem,
        llm: LLM,
        solution: Solution,
        log_dir="",
        seed=None,
    ):
        """
        Normally called at the end of a run.
        1) Logs final run metadata to MLflow
        2) Ends the MLflow run
        3) Calls super().add_run(...) so we keep the file-based logs
        """
        # --- MLflow logging ---
        if not self._mlflow_run_active:
            # If there's no open run, you could open one automatically
            # or simply warn. We'll open one just in case:
            run_name = f"{method.name}_{problem.name}_seed{seed}"
            run = mlflow.start_run(experiment_id=self.experiment_id, run_name=run_name)
            self._mlflow_run_active = True
            self._current_run_id = run.info.run_id

        # Log whatever final data you like
        mlflow.log_param("method_name", method.name)
        mlflow.log_param("problem_name", problem.name)
        mlflow.log_param("llm_name", llm.model)
        if seed is not None:
            mlflow.log_param("seed", seed)

        # Log final fitness as a metric
        final_fitness = (
            solution.fitness if solution.fitness is not None else float("nan")
        )
        mlflow.log_metric("final_fitness", final_fitness)

        # Log a serialized object as an artifact
        rel_log_dir = os.path.relpath(log_dir, self.dirname)
        run_object = {
            "method_name": method.name,
            "problem_name": problem.name,
            "llm_name": llm.model,
            "method": method.to_dict(),
            "problem": problem.to_dict(),
            "llm": llm.to_dict(),
            "solution": solution.to_dict(),
            "log_dir": rel_log_dir,
            "seed": seed,
        }
        mlflow.log_text(
            json.dumps(convert_to_serializable(run_object), indent=2),
            artifact_file="final_run_object.json",
        )

        # End the MLflow run
        mlflow.end_run()
        self._mlflow_run_active = False
        self._current_run_id = None

        # --- File-based logging (super) ---
        super().add_run(
            method=method,
            problem=problem,
            llm=llm,
            solution=solution,
            log_dir=log_dir,
            seed=seed,
        )

    # Optionally override get_data and get_problem_data to *also*
    # combine data from MLflow runs. Right now, we leave them alone
    # so they read from your local experimentlog.jsonl as before.


class MLFlowRunLogger(RunLogger):
    """
    A RunLogger subclass that logs data to MLflow *and* to file,
    relying on the fact that the MLFlowExperimentLogger has opened a run.
    """

    def __init__(self, name="", root_dir="", budget=100, progress_callback=None):
        # We do want to keep the parent's file-based logging directories
        super().__init__(name, root_dir, budget, progress_callback=progress_callback)

    def log_conversation(self, role, content, cost=0.0, tokens=0):
        """
        Logs conversation details to MLflow, plus calls super() to keep the local file logs if you wish.
        """
        # Log to MLflow (assumes an MLflow run is active)
        conversation = {
            "role": role,
            "time": str(datetime.now()),
            "content": content,
            "cost": float(cost),
            "tokens": int(tokens),
        }
        # Example: We can log each conversation snippet as a text artifact
        mlflow.log_text(
            json.dumps(conversation), artifact_file=f"conversation_{self.attempt}.json"
        )
        self.attempt += 1

        # Also do the usual file-based logs or any other logic you have in the parent
        # (You may have to define such a method if you want to keep conversation logs in a file.)
        super().log_conversation(role, content, cost, tokens)

    def log_individual(self, individual):
        """
        Logs an individual solution to MLflow, then calls super() to keep the normal file logging.
        """
        ind_dict = individual.to_dict()

        # Example: log a metric if there's a fitness
        if "fitness" in ind_dict:
            mlflow.log_metric("fitness", ind_dict["fitness"])

        # Store the entire solution as a JSON artifact
        mlflow.log_text(
            json.dumps(convert_to_serializable(ind_dict)),
            artifact_file=f"solution_{individual.id}.json",
        )

        # Keep file-based logging
        super().log_individual(individual)

    def log_code(self, individual):
        """
        Log code as a text artifact in MLflow, plus the normal .py file in your local run folder.
        """
        mlflow.log_text(individual.code, artifact_file=f"code_{individual.id}.py")
        super().log_code(individual)

    def log_configspace(self, individual):
        """
        If there's a config space, log it to MLflow as well as the local file.
        """
        if individual.configspace is not None:
            cs_text = cs_json.write(individual.configspace)
        else:
            cs_text = "Failed to extract config space"
        mlflow.log_text(cs_text, artifact_file=f"configspace_{individual.id}.json")

        super().log_configspace(individual)

    def budget_exhausted(self):
        """
        Optionally still rely on the file-based approach for counting lines in log.jsonl
        or store a separate counter. For now, we call super() to preserve the old logic.
        """
        return super().budget_exhausted()



from __future__ import annotations

import logging
import traceback
from types import SimpleNamespace
from typing import Any

from ..llm import LLM
from ..method import Method
from ..problem import Problem
from ..solution import Solution
from ..utils import class_info, first_class_name

try:
    from reevo import ReEvo as ReEvoAlgorithm
    from reevo.utils.llm_client.base import BaseClient
except Exception:  # pragma: no cover - optional dependency
    ReEvoAlgorithm = None
    BaseClient = object  # type: ignore


class _BladeReEvoClient(BaseClient):
    """Adapter that exposes the interface expected by ReEvo."""

    def __init__(self, llm: LLM, temperature: float = 1.0) -> None:
        super().__init__(model=llm.model, temperature=temperature)
        self.llm = llm

    def _chat_completion_api(
        self, messages: list[dict], temperature: float, n: int = 1
    ):
        responses = []
        for _ in range(n):
            content = self.llm.query(messages)
            responses.append({"message": {"role": "assistant", "content": content}})
        return responses

    def multi_chat_completion(
        self,
        messages_list,
        n=1,
        temperature=None,
    ):
        """Sequentially generate responses for many independent conversations.

        Parameters
        ----------
        messages_list : list[list[dict]] | list[dict]
            Either *one* conversation (list[dict]) or a list of conversations.
        n : int, default=1
            Number of completions **per conversation**.  For multiple
            conversations `n` must remain 1.
        temperature : float | None, default=None
            Sampling temperature.

        Returns
        -------
        list[str]
            The content field of each returned message, flattened.
        """
        # Normalise input ---------------------------------------------------
        if not isinstance(messages_list, list):
            raise TypeError("messages_list must be a list")
        if messages_list and not isinstance(messages_list[0], list):
            # Single conversation – wrap it so downstream code can iterate.
            messages_list = [messages_list]  # type: ignore[assignment]

        if len(messages_list) > 1 and n != 1:
            raise ValueError(
                "Currently, only n=1 is supported for multi‑chat completion."
            )

        # ------------------------------------------------------------------
        # Sequential execution (no ThreadPoolExecutor) ----------------------
        # ------------------------------------------------------------------
        contents = []
        for msgs in messages_list:
            for choice in self.chat_completion(
                n=n, messages=msgs, temperature=temperature
            ):
                contents.append(choice["message"]["content"])
        return contents


class ReEvo(Method):
    """Wrapper for the ReEvo baseline."""

    def __init__(self, llm: LLM, budget: int, name: str = "ReEvo", **kwargs: Any):
        super().__init__(llm, budget, name)
        self.kwargs = kwargs

    def _eval_population(self, reevo: Any, population: list[dict], problem: Problem):
        for response_id in range(len(population)):
            individual = population[response_id]
            reevo.function_evals += 1
            if individual.get("code") is None:
                individual["exec_success"] = False
                individual["obj"] = float("inf")
                continue
            solution = Solution(
                code=individual["code"],
                name=first_class_name(individual["code"]) or "AlgorithmName",
                description=class_info(individual["code"])[1]
                or "No description provided.",
            )
            solution = problem(solution)

            if solution.error != "":
                # If the solution has an error, we mark it as invalid.
                individual["exec_success"] = False
                individual["obj"] = -solution.fitness
                population[response_id] = reevo.mark_invalid_individual(
                    individual, solution.error
                )
                continue
            # Re-Evo always minimizes. (while BLADE problems are maximization)
            individual["obj"] = -solution.fitness

            individual["exec_success"] = True
            population[response_id] = individual
        return population

    def __call__(self, problem: Problem):
        if ReEvoAlgorithm is None:
            raise ImportError(
                "reevo package is not installed, please install it using `poetry install --with methods`."
            )

        from omegaconf import OmegaConf

        cfg_dict = {
            "max_fe": self.budget,
            "pop_size": self.kwargs.get("pop_size", 10),
            "init_pop_size": self.kwargs.get("init_pop_size", 20),
            "mutation_rate": self.kwargs.get("mutation_rate", 0.5),
            "timeout": self.kwargs.get("timeout", 20),
            "problem": {
                "problem_name": problem.name,
                "description": problem.task_prompt,
                "problem_size": getattr(problem, "dim", 1),
                "func_name": "AlgorithmName",
                "seed_func": problem.example_prompt,
                "func_signature": f"{problem.func_name}(self)",
                "obj_type": "max",
                "problem_type": "blade",
                "func_desc": "",
                "external_knowledge": "",
            },
        }
        cfg = OmegaConf.create(cfg_dict)
        client = _BladeReEvoClient(self.llm)
        reevo = ReEvoAlgorithm(
            cfg, root_dir=self.kwargs.get("output_path", "./"), generator_llm=client
        )
        # Override evaluation to use BLADE problems
        reevo.evaluate_population = lambda pop: self._eval_population(
            reevo, pop, problem
        )
        reevo.init_population()
        code, _ = reevo.evolve()
        name = first_class_name(code) or "AlgorithmName"
        sol = Solution(code=code, name=name)
        sol.set_scores(-reevo.best_obj_overall, "", "")
        return sol

    def to_dict(self):
        return {
            "method_name": self.name if self.name is not None else "ReEvo",
            "budget": self.budget,
            "kwargs": self.kwargs,
        }



import ast
import difflib
import json
import re
import textwrap
from typing import Optional, Tuple

import numpy as np
from ioh import LogInfo, logger


class TimeoutException(Exception):
    """Custom exception for handling timeouts."""

    pass


def first_class_name(code_string: str) -> str | None:
    try:  # 1. do it the robust way
        for node in ast.parse(code_string).body:  #    (won’t be fooled by comments)
            if isinstance(node, ast.ClassDef):
                return node.name
    except SyntaxError:
        pass  # fall back if the snippet is malformed

    m = re.search(
        r"^\s*class\s+([A-Za-z_]\w*)\s*[:(]", code_string, re.M
    )  # 2. quick-n-dirty
    return m.group(1) if m else None


def class_info(code_string: str) -> Tuple[Optional[str], Optional[str]]:
    """
    Return (class_name, docstring_or_None).
    """
    # 1.  Robust path: use the AST – ignores comments and embedded strings
    try:
        for node in ast.parse(code_string).body:
            if isinstance(node, ast.ClassDef):
                return node.name, ast.get_docstring(node)
    except SyntaxError:
        pass  # malformed snippet ⇒ fall back

    # 2.  Fallback: quick regex for the first  class …\n  "docstring"
    m = re.search(
        r"^\s*class\s+([A-Za-z_]\w*)\s*[:(][^\n]*\n"  # class line
        r'\s*(?P<q>["\']){1,3}(?P<doc>.*?)(?P=q){1,3}',  # first quoted block/line
        code_string,
        re.S | re.M,
    )
    return (m.group(1), textwrap.dedent(m.group("doc")).strip()) if m else (None, None)


def code_compare(code1, code2):
    # Parse the Python code into ASTs
    # Use difflib to find differences
    diff = difflib.ndiff(code1.splitlines(), code2.splitlines())
    # Count the number of differing lines
    diffs = sum(1 for x in diff if x.startswith("- ") or x.startswith("+ "))
    # Calculate total lines for the ratio
    total_lines = max(len(code1.splitlines()), len(code2.splitlines()))
    similarity_ratio = (total_lines - diffs) / total_lines if total_lines else 1
    return 1 - similarity_ratio


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


def convert_to_serializable(data):
    if isinstance(data, dict):
        return {key: convert_to_serializable(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [convert_to_serializable(item) for item in data]
    elif isinstance(data, np.integer):
        return int(data)
    elif isinstance(data, np.floating):
        return float(data)
    if isinstance(data, np.ndarray):
        return data.tolist()
    else:
        if is_jsonable(data):
            return data
        else:
            return str(data)


class NoCodeException(Exception):
    """Could not extract generated code."""

    pass


class ThresholdReachedException(Exception):
    """The algorithm reached the lower threshold."""

    pass


class OverBudgetException(Exception):
    """The algorithm tried to do more evaluations than allowed."""

    pass


def correct_aoc(ioh_function, logger, budget):
    """Correct aoc values in case a run stopped before the budget was exhausted

    Args:
        ioh_function: The function in its final state (before resetting!)
        logger: The logger in its final state, so we can ensure the settings for aoc calculation match
        budget: The intended maximum budget

    Returns:
        float: The normalized aoc of the run, corrected for stopped runs
    """
    fraction = (
        logger.transform(
            np.clip(
                ioh_function.state.current_best_internal.y, logger.lower, logger.upper
            )
        )
        - logger.transform(logger.lower)
    ) / (logger.transform(logger.upper) - logger.transform(logger.lower))
    aoc = (
        logger.aoc
        + np.clip(budget - ioh_function.state.evaluations, 0, budget) * fraction
    ) / budget

    return 1 - aoc


class aoc_logger(logger.AbstractLogger):
    """aoc_logger class implementing the logging module for ioh."""

    def __init__(
        self,
        budget,
        lower=1e-8,
        upper=1e8,
        scale_log=True,
        stop_on_threshold=False,
        *args,
        **kwargs,
    ):
        """Initialize the logger.

        Args:
            budget (int): Evaluation budget for calculating aoc.
        """
        super().__init__(*args, **kwargs)
        self.aoc = 0
        self.lower = lower
        self.upper = upper
        self.budget = budget
        self.stop_on_threshold = stop_on_threshold
        self.transform = lambda x: np.log10(x) if scale_log else (lambda x: x)

    def __call__(self, log_info: LogInfo):
        """Subscalculate the aoc.

        Args:
            log_info (ioh.LogInfo): info about current values.
        """
        if log_info.evaluations > self.budget:
            raise OverBudgetException
        if log_info.evaluations == self.budget:
            return
        if self.stop_on_threshold and abs(log_info.raw_y_best) < self.lower:
            raise ThresholdReachedException
        y_value = np.clip(log_info.raw_y_best, self.lower, self.upper)
        self.aoc += (self.transform(y_value) - self.transform(self.lower)) / (
            self.transform(self.upper) - self.transform(self.lower)
        )

    def reset(self, func):
        super().reset()
        self.aoc = 0


class budget_logger(logger.AbstractLogger):
    """budget_logger class implementing the logging module for ioh."""

    def __init__(
        self,
        budget,
        *args,
        **kwargs,
    ):
        """Initialize the logger.

        Args:
            budget (int): Evaluation budget for calculating aoc.
        """
        super().__init__(*args, **kwargs)
        self.budget = budget

    def __call__(self, log_info: LogInfo):
        """Subscalculate the aoc.

        Args:
            log_info (ioh.LogInfo): info about current values.
        """
        if log_info.evaluations > self.budget:
            raise OverBudgetException

    def reset(self):
        super().reset()



from iohblade.experiment import Experiment
from iohblade.llm import Gemini_LLM, Ollama_LLM, OpenAI_LLM
from iohblade.methods import LLaMEA, RandomSearch, EoH, ReEvo
from iohblade.loggers import ExperimentLogger
from iohblade.problems import AutoML
import numpy as np
import os
import logging

if __name__ == "__main__": # prevents weird restarting behaviour
    api_key = os.getenv("OPENAI_API_KEY")
    api_key_gemini = os.getenv("GEMINI_API_KEY")

    #llm1 = Gemini_LLM(api_key_gemini, "gemini-1.5-flash")
    llm2 = OpenAI_LLM(api_key,"gpt-4.1-nano-2025-04-14", temperature=1.0)
    budget = 10

    mutation_prompts = [
        "Refine the strategy of the selected solution to improve it.",  # small mutation
        "Generate a new algorithm that is different from the algorithms you have tried before.", #new random solution
        "Refine and simplify the selected algorithm to improve it.", #simplify
    ]

    for llm in [llm2]:
        LLaMEA_method = LLaMEA(llm, budget=budget, name="LLaMEA", mutation_prompts=mutation_prompts, n_parents=2, n_offspring=4, elitism=True)
        ReEvo_method = ReEvo(llm, budget=budget, name="ReEvo", output_path="results/automl-breast-cancer", pop_size=2, init_pop_size=4)
        EoH_method = EoH(llm, budget=budget, name="EoH", output_path="results/automl-breast-cancer")
        methods = [EoH_method, ReEvo_method, LLaMEA_method] 
        logger = ExperimentLogger("results/automl-breast-cancer")
        problems = [AutoML()]
        experiment = Experiment(methods=methods, problems=problems, runs=1, show_stdout=False, exp_logger=logger, budget=budget, n_jobs=3) #normal run

        experiment() #run the experiment


import ast
import json
import logging
import os
import re
import textwrap
from typing import Optional, Tuple

from ..llm import LLM
from ..method import Method
from ..problem import Problem
from ..solution import Solution
from ..utils import class_info, first_class_name

try:
    from eoh import eoh as eoh_main
    from eoh.methods.eoh import eoh_evolution
    from eoh.utils.getParas import Paras
except Exception:  # pragma: no cover - optional dependency
    eoh_main = None
    Paras = None
    eoh_evolution = None


class _BladePrompts:
    """Adapter exposing the prompt interface expected by EoH."""

    def __init__(self, problem: Problem):
        self.problem = problem

    def get_task(self):
        return self.problem.task_prompt

    def get_func_name(self):
        return self.problem.func_name

    def get_func_inputs(self):
        return self.problem.func_inputs

    def get_init_inputs(self):
        return self.problem.init_inputs

    def get_func_outputs(self):
        return self.problem.func_outputs

    def get_inout_inf(self):
        return (
            f"Implement a Python class called `AlgorithmName` with"
            f" an __init__(self, {', '.join(self.get_init_inputs())}) and a function {self.get_func_name()}(self, {', '.join(self.get_func_inputs())})"
            f" returning {', '.join(self.get_func_outputs())}."
        )

    def get_other_inf(self):
        return (
            self.problem.task_prompt
            + self.problem.example_prompt
            + self.problem.format_prompt
        )


class _BladeProblemAdapter:
    """Wraps a BLADE problem for use with EoH."""

    def __init__(self, problem: Problem):
        self.problem = problem
        self.prompts = _BladePrompts(problem)

    def evaluate(self, code_string):
        solution = Solution(
            code=code_string,
            name=first_class_name(code_string) or "AlgorithmName",
            description=class_info(code_string)[1] or "No description provided.",
        )
        solution = self.problem(solution)
        return (
            -solution.fitness
        )  # EoH minimizes the fitness, so we return negative value.


class _BladeInterfaceLLM:
    """Provide EoH with the BLADE LLM interface."""

    def __init__(self, llm: LLM):
        self.llm = llm

    def get_response(self, prompt_content):
        return self.llm.query([{"role": "user", "content": prompt_content}])


class EoH(Method):
    def __init__(self, llm: LLM, budget: int, name="EoH", **kwargs):
        """
        Initializes the EoH algorithm within the benchmarking framework.

        Args:
            problem (Problem): The problem instance to optimize.
            llm (LLM): The LLM instance to use for solution generation.
            budget (int): The maximum number of evaluations.
            name (str): The name of the method.
            kwargs: Additional arguments for configuring EoH.
        """
        super().__init__(llm, budget, name)
        self.kwargs = kwargs

    def __call__(self, problem: Problem):
        """
        Executes the evolutionary search process via EoH.

        Returns:
            Solution: The best solution found.
        """
        if eoh_main is None:
            raise ImportError(
                "EoH package is not installed, , please install it using `poetry install --with methods`."
            )

        # Patch the EoH LLM interface to use our LLM
        if eoh_evolution is not None:
            eoh_evolution.InterfaceLLM = lambda *args, **kwargs: _BladeInterfaceLLM(
                self.llm
            )

        paras = Paras()
        paras.set_paras(
            method="eoh",
            problem=_BladeProblemAdapter(problem),
            llm_api_endpoint="unused",
            llm_api_key="unused",
            llm_model=self.llm.model,
            ec_pop_size=self.kwargs.get("pop_size", 4),
            ec_n_pop=max(1, self.budget // self.kwargs.get("pop_size", 4)),
            exp_output_path=self.kwargs.get("output_path", "./"),
            exp_n_proc=1,
            exp_debug_mode=False,
        )

        evol = eoh_main.EVOL(paras, prob=paras.problem)
        evol.run()

        result_file = os.path.join(
            paras.exp_output_path,
            "results",
            "pops_best",
            f"population_generation_{paras.ec_n_pop}.json",
        )
        if not os.path.exists(result_file):
            raise FileNotFoundError(result_file)

        with open(result_file) as f:
            best = json.load(f)

        if isinstance(best, list):
            best = best[0]

        code = best.get("code", "")
        desc = best.get("algorithm", "")
        name_match = re.search(r"class\s+(\w+)", code)
        name = name_match.group(1) if name_match else "OptimizationAlgorithm"
        solution = Solution(code=code, name=name, description=desc)
        solution.set_scores(-best.get("objective", 0))
        return solution

    def to_dict(self):
        """
        Returns a dictionary representation of the method including all parameters.

        Returns:
            dict: Dictionary representation of the method.
        """
        return {
            "method_name": self.name if self.name != None else "EoH",
            "budget": self.budget,
            "kwargs": self.kwargs,
        }



import ioh
import numpy as np
import pandas as pd
import pytest

import iohblade.behaviour_metrics as bm
import iohblade.llm as llm_mod
import iohblade.utils as utils


@pytest.fixture
def sample_df():
    return pd.DataFrame(
        {
            "evaluations": [1, 2, 3, 4],
            "raw_y": [4.0, 2.0, 3.0, 1.0],
            "x0": [0.0, 1.0, 2.0, 0.0],
            "x1": [0.0, 0.0, 1.0, 0.0],
        }
    )


def test_individual_metrics(sample_df):
    df = sample_df
    assert bm.get_coordinates(df).shape == (4, 2)
    assert bm.get_objective(df).tolist() == [4.0, 2.0, 3.0, 1.0]
    assert bm.average_nearest_neighbor_distance(df, step=1) == pytest.approx(0.80473785)
    bounds = [(0.0, 2.0), (0.0, 1.0)]
    rng = np.random.default_rng(0)
    assert bm.coverage_dispersion(df, bounds=bounds, n_samples=5, rng=rng) >= 0
    assert bm.spatial_entropy(df) > 0
    assert bm.average_distance_to_best_so_far(df) == pytest.approx(0.47140452)
    assert bm.closed_form_random_search_diversity(bounds) == pytest.approx(0.86602540)
    rng = np.random.default_rng(0)
    assert bm.estimate_random_search_diversity(bounds, n=3, samples=5, rng=rng) > 0
    exp, exl = bm.avg_exploration_exploitation_chunked(df, chunk_size=2, bounds=bounds)
    assert exp == 100.0 and exl == 0.0
    assert bm.intensification_ratio(df, radius=1.5) == pytest.approx(0.75)
    assert bm.average_convergence_rate(df, optimum=0) < 1
    avg_imp, succ = bm.improvement_statistics(df)
    assert avg_imp == pytest.approx(1.5)
    assert succ == pytest.approx(2 / 3)
    assert bm.longest_no_improvement_streak(df) == 1
    assert bm.last_improvement_fraction(df) == 0.0


def test_compute_behavior_metrics_deterministic(sample_df, monkeypatch):
    df = sample_df
    original = np.random.default_rng

    def fixed_rng(_seed=None):
        return original(0)

    monkeypatch.setattr(np.random, "default_rng", fixed_rng)

    metrics = bm.compute_behavior_metrics(
        df, bounds=[(0.0, 2.0), (0.0, 1.0)], radius=1.5, disp_samples=5
    )

    expected_keys = {
        "avg_nearest_neighbor_distance",
        "dispersion",
        "avg_exploration_pct",
        "avg_distance_to_best",
        "intensification_ratio",
        "avg_exploitation_pct",
        "average_convergence_rate",
        "avg_improvement",
        "success_rate",
        "longest_no_improvement_streak",
        "last_improvement_fraction",
    }
    assert expected_keys <= metrics.keys()
    assert metrics["avg_improvement"] == pytest.approx(1.5)
    assert metrics["success_rate"] == pytest.approx(2 / 3)


def test_metric_edge_cases(monkeypatch):
    df_single = pd.DataFrame({"evaluations": [1], "raw_y": [1.0], "x0": [0.0]})
    df_two = pd.DataFrame(
        {"evaluations": [1, 2], "raw_y": [1.0, 0.5], "x0": [0.0, 0.1]}
    )

    # _pairwise_distances should return empty array for single point
    X = bm.get_coordinates(df_single)
    assert bm._pairwise_distances(X).size == 0

    # average_nearest_neighbor_distance should early exit
    assert bm.average_nearest_neighbor_distance(df_single) == 0.0

    # coverage_dispersion with default bounds
    disp = bm.coverage_dispersion(
        df_single, bounds=None, n_samples=5, rng=np.random.default_rng(0)
    )
    assert disp >= 0

    # estimate_random_search_diversity with default rng
    est = bm.estimate_random_search_diversity([(0.0, 1.0)], n=2, samples=3)
    assert est >= 0

    # avg_exploration_exploitation_chunked should return defaults when not enough points
    expl, explo = bm.avg_exploration_exploitation_chunked(df_single, chunk_size=2)
    assert (expl, explo) == (0.0, 100.0)

    # compute_behavior_metrics with defaults using two points
    metrics = bm.compute_behavior_metrics(df_two)
    assert "avg_nearest_neighbor_distance" in metrics


class DummyLogger:
    def __init__(self):
        self.logged = []

    def log_conversation(self, who, text, cost=0.0, tokens=0):
        self.logged.append((who, text, cost, tokens))


class DummyLLM(llm_mod.LLM):
    def __init__(self):
        super().__init__(api_key="", model="dummy", logger=DummyLogger())

    def _query(self, session_messages):
        return """```python\nclass Algo:\n    pass\n```\n# Description: test"""


def test_sample_solution_runs():
    llm = DummyLLM()
    solution = llm.sample_solution([{"role": "user", "content": "hello"}])
    assert solution.name == "Algo"
    assert "pass" in solution.code
    assert solution.description == "test"
    assert llm.to_dict()["model"] == "dummy"


def test_aoc_budget_logger_updates():
    logger_instance = utils.aoc_logger(
        budget=2, upper=1e2, triggers=[ioh.logger.trigger.ALWAYS]
    )
    info = ioh.LogInfo(
        evaluations=1,
        raw_y=5.0,
        raw_y_best=5.0,
        transformed_y=0.0,
        transformed_y_best=0.0,
        y=5.0,
        y_best=5.0,
        x=[0.0],
        violations=[],
        penalties=[],
        optimum=ioh.iohcpp.RealSolution([1.0], -1.0),
        has_improved=False,
    )
    logger_instance(info)
    assert logger_instance.aoc > 0
    logger_instance.reset(None)
    assert logger_instance.aoc == 0

    budget = utils.budget_logger(budget=1, triggers=[ioh.logger.trigger.ALWAYS])
    with pytest.raises(utils.OverBudgetException):
        budget(
            ioh.LogInfo(
                evaluations=2,
                raw_y=0.0,
                raw_y_best=0.0,
                transformed_y=0.0,
                transformed_y_best=0.0,
                y=0.0,
                y_best=0.0,
                x=[0.0],
                violations=[],
                penalties=[],
                optimum=ioh.iohcpp.RealSolution([1.0], -1.0),
                has_improved=False,
            )
        )



import os
import traceback

import ioh
import numpy as np
import pandas as pd
from ConfigSpace import Configuration, ConfigurationSpace
from ioh import get_problem
from ioh import logger as ioh_logger
from smac import AlgorithmConfigurationFacade, Scenario

from ..problem import Problem
from ..solution import Solution
from ..utils import OverBudgetException, aoc_logger, correct_aoc


class BBOB_SBOX(Problem):
    """
    Problem class for evaluating optimization algorithms on the SBOX-COST or BBOB benchmark. See also https://github.com/sbox-cost/Examples

    Black-Box Optimization Benchmarking (BBOB) problem set, which contains 24 noiseless real-valued test functions supported on [-5, 5]^n, where n is the dimensionality.

    These problems were orginally proposed by Hansen et. al. in [FinckHRA10] and was implemented as the core component of the COmparing Continous Optimizer (COCO) platform [HansenARMTB20].
    We took the implementation of those 24 functions in https://github.com/numbbo/coco/tree/master/code-experiments/src (v2.2) and adopted those to our framework.


    [HansenARMTB20] Nikolaus Hansen, Anne Auger, Raymond Ros, Olaf Mersmann, Tea Tusar, and Dimo Brockhoff. “COCO: A platform for comparing continuous optimizers in a black-box setting.” Optimization Methods and Software (2020): 1-31.

    [FinckHRA10] Steffen Finck, Nikolaus Hansen, Raymond Ros, and Anne Auger. “Real-parameter black-box optimization benchmarking 2009: Presentation of the noiseless functions.” Technical Report 2009/20, Research Center PPE, 2009. Updated February, 2010.
    """

    def __init__(
        self,
        logger=None,
        training_instances=None,
        test_instances=None,
        name="SBOX_COST",
        eval_timeout=120,
        dims=[2, 5],
        budget_factor=2000,
        specific_fid=None,
        specific_group=None,
        problem_type=ioh.ProblemClass.SBOX,
        full_ioh_log=False,
        ioh_dir="",
    ):
        """
        Initializes the MA-BBOB problem instance.
        Args:
            logger (RunLogger): The logger to use for logging.
            training_instances (list): A list of tuples with (fid=function id, iid=instance id) for training instances to use.
            test_instances (list): The indices of test instances to use. A list of tuples with (fid=function id, iid=instance id).
            name (str): The name of the problem.
            eval_timeout (int): The evaluation timeout in seconds.
            dims (list): The dimensionalities of the problem instances to run on.
            budget_factor (int): The factor to multiply the dimensionality with to get the budget.
            specific_fid (int): The specific function id to use. If not None, additional information is added to the prompt about the function.
            specific_group (int): The specific function group (1,2,3,4,5) to use. If not None, additional information is added to the prompt about the function group.
            problem_type (ioh.ProblemClass): The type of problem to use. Can be SBOX or BBOB.
            full_ioh_log (bool): If set to True, additional IOH logs are being kept for each run and each algorithm.
        """
        if training_instances is None:
            training_instances = [(f, i) for f in range(1, 25) for i in range(1, 6)]
        if test_instances is None:
            test_instances = [
                (f, i) for f in range(1, 25) for i in range(5, 16)
            ]  # 10 test instances
        super().__init__(logger, training_instances, test_instances, name, eval_timeout)
        self.dims = dims  # The dimensionalities of the problem instances to run on
        self.budget_factor = budget_factor  # The factor to multiply the dimensionality with to get the budget
        self.specific_fid = specific_fid
        self.specific_group = specific_group
        self.full_ioh_log = full_ioh_log
        self.ioh_dir = ioh_dir

        self.func_name = "__call__"
        self.init_inputs = ["budget", "dim"]
        self.func_inputs = ["func"]
        self.func_outputs = ["f_opt", "x_opt"]

        # List containing descriptions of each function group
        function_groups = [
            "Separable Functions",
            "Functions with low or moderate conditioning",
            "Functions with high conditioning and unimodal",
            "Multi-modal functions with adequate global structure",
            "Multi-modal functions with weak global structure",
        ]

        # List containing information for all 24 functions
        functions = [
            "f1: Sphere Function",
            "f2: Separable Ellipsoidal Function",
            "f3: Rastrigin Function",
            "f4: Büche-Rastrigin Function",
            "f5: Linear Slope",
            "f6: Attractive Sector Function",
            "f7: Step Ellipsoidal Function",
            "f8: Rosenbrock Function, original",
            "f9: Rosenbrock Function, rotated",
            "f10: Ellipsoidal Function",
            "f11: Discus Function",
            "f12: Bent Cigar Function",
            "f13: Sharp Ridge Function",
            "f14: Different Powers Function",
            "f15: Rastrigin Function",
            "f16: Weierstrass Function",
            "f17: Schaffer's F7 Function",
            "f18: Schaffer's F7 Function, moderately ill-conditioned",
            "f19: Composite Griewank-Rosenbrock Function F8F2",
            "f20: Schwefel Function",
            "f21: Gallagher's Gaussian 101-me Peaks Function",
            "f22: Gallagher's Gaussian 21-hi Peaks Function",
            "f23: Katsuura Function",
            "f24: Lunacek bi-Rastrigin Function",
        ]
        self.problem_type = problem_type
        self.benchmark_name = (
            "SBOX-COST test suite of noiseless box-constrained functions."
            if problem_type == ioh.ProblemClass.SBOX
            else "BBOB test suite of noiseless functions."
        )
        box_constrained = (
            "box-constrained"
            if problem_type == ioh.ProblemClass.SBOX
            else "unconstrained"
        )
        extra_prompt = f"The optimization algorithm should handle a wide range of tasks, which is evaluated on the {self.benchmark_name}"
        if (
            self.specific_fid is not None
            and self.specific_fid < 25
            and self.specific_fid > 0
        ):
            extra_prompt = f"The optimization algorithm should work on different instances of noiseless {box_constrained} functions. Specifically function: {functions[self.specific_fid-1]}."
        elif (
            self.specific_group is not None
            and self.specific_group < 6
            and self.specific_group > 0
        ):
            extra_prompt = f"The optimization algorithm should work on different instances of noiseless {box_constrained} functions. Specifically it should work well for {function_groups[self.specific_group-1]}."
        else:
            extra_prompt = f"The optimization algorithm should work on different instances of noiseless {box_constrained} functions."

        self.task_prompt = f"""
You are a Python expert working on a new optimization algorithm.
Your task is to develop a novel heuristic optimization algorithm for continuous optimization problems.
{extra_prompt} Your task is to write the optimization algorithm in Python code. 
Each of the optimization functions has a search space between -5.0 (lower bound) and 5.0 (upper bound). The dimensionality can be varied.
The code should contain an `__init__(self, budget, dim)` function with optional additional arguments and the function `def __call__(self, func)`, which should optimize the black box function `func` using `self.budget` function evaluations.
The func() can only be called as many times as the budget allows, not more. 
"""
        self.example_prompt = """
An example of such code (a simple random search), is as follows:
```python
import numpy as np

class RandomSearch:
    def __init__(self, budget=10000, dim=10):
        self.budget = budget
        self.dim = dim

    def __call__(self, func):
        self.f_opt = np.Inf
        self.x_opt = None
        for i in range(self.budget):
            x = np.random.uniform(func.bounds.lb, func.bounds.ub)
            
            f = func(x)
            if f < self.f_opt:
                self.f_opt = f
                self.x_opt = x
            
        return self.f_opt, self.x_opt
```
        """
        self.format_prompt = """
Give an excellent and novel heuristic algorithm to solve this task and also give it a one-line description, describing the main idea. Give the response in the format:
# Description: <short-description>
# Code: 
```python
<code>
```
"""

    def get_prompt(self):
        """
        Returns the problem description and answer format.
        """
        return self.task_prompt + self.example_prompt + self.format_prompt

    def evaluate(self, solution: Solution, test=False):
        """
        Evaluates a solution on the SBOX or BBOB benchmark using AOCC.
        """
        auc_mean = 0
        auc_std = 0
        code = solution.code
        algorithm_name = solution.name
        algorithm_id = solution.id
        safe_globals = {"np": np}
        local_env = {}
        exec(code, safe_globals, local_env)

        algorithm = None

        # Small test run to catch code errors
        try:
            l2_temp = aoc_logger(100, upper=1e2, triggers=[ioh_logger.trigger.ALWAYS])
            problem = get_problem(
                1, instance=1, dimension=2, problem_class=self.problem_type
            )
            problem.attach_logger(l2_temp)
            algorithm = local_env[algorithm_name](budget=100, dim=2)
            algorithm(problem)
        except OverBudgetException:
            pass

        # Final validation
        instances = self.test_instances if test else self.training_instances
        aucs = []
        for dim in self.dims:
            for instance in instances:
                fid, iid = instance  # we expact a tuple of (fid, iid)
                budget = self.budget_factor * dim
                f_new = get_problem(
                    fid, instance=iid, dimension=dim, problem_class=self.problem_type
                )
                l2 = aoc_logger(budget, upper=1e2, triggers=[ioh_logger.trigger.ALWAYS])
                if test or self.full_ioh_log:
                    l1 = ioh.logger.Analyzer(
                        root=self.ioh_dir,
                        folder_name=algorithm_id,
                        algorithm_name=algorithm_id,
                        store_positions=True,
                        triggers=[ioh_logger.trigger.ALWAYS],
                    )
                    combined_logger = ioh.logger.Combine([l1, l2])
                    f_new.attach_logger(combined_logger)
                else:
                    f_new.attach_logger(l2)

                try:
                    algorithm = local_env[algorithm_name](budget=budget, dim=dim)
                    algorithm(f_new)
                except OverBudgetException:
                    pass

                aucs.append(correct_aoc(f_new, l2, budget))
                l2.reset(f_new)
                f_new.reset()

        auc_mean = np.mean(aucs)
        auc_std = np.std(aucs)

        solution.add_metadata("aucs", aucs)
        solution.set_scores(
            auc_mean,
            f"The algorithm {algorithm_name} scored {auc_mean:.3f} on AOCC (higher is better, 1.0 is the best).",
        )

        return solution

    def test(self, solution: Solution):
        """
        Runs the solution on test instances and returns the fitness score.
        """
        return self.evaluate(solution, True)

    def to_dict(self):
        """
        Converts the problem to a dictionary.
        """
        return {
            "name": self.name,
            "dims": self.dims,
            "training_instances": self.training_instances,
            "test_instances": self.test_instances,
            "budget_factor": self.budget_factor,
            "problem_type": "SBOX"
            if self.problem_type == ioh.ProblemClass.SBOX
            else "BBOB",
            "specific_fid": self.specific_fid,
            "specific_group": self.specific_group,
        }



import iohinspector
import polars as pl
from iohblade.behaviour_metrics import compute_behavior_metrics
from tqdm import tqdm
import pandas as pd
import numpy as np
from iohblade.loggers import ExperimentLogger
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
import seaborn as sns
from pandas.plotting import parallel_coordinates
from iohblade import plot_convergence, plot_experiment_CEG, plot_boxplot_fitness_hue, plot_boxplot_fitness, fitness_table
import os
import tqdm
from sklearn.preprocessing import RobustScaler, MinMaxScaler


# ── 2. HELPER FUNCTION (drop it right after `behaviour_feats` is first defined) ─
def robust_minmax(df, cols, q=(0.01, 0.99)):
    """
    Robust [0-1] scaling that flattens outliers.

    • Clips each feature to the q-th and (1-q)-th quantile.  
    • Then rescales the clipped values to the 0-1 interval.
    """
    lo = df[cols].quantile(q[0])
    hi = df[cols].quantile(q[1])
    df[cols] = df[cols].clip(lo, hi, axis=1)
    return (df[cols] - lo) / (hi - lo)

"""
    mutation_prompts1 = [
        "Refine and simplify the selected algorithm to improve it.",  # simplify mutation
    ]
    mutation_prompts2 = [
        "Generate a new algorithm that is different from the algorithms you have tried before.", #new random solution
    ]
    mutation_prompts3 = [
        "Refine and simplify the selected solution to improve it.",  # simplify mutation
        "Generate a new algorithm that is different from the algorithms you have tried before.", #new random solution
    ]
    LLaMEA_method1 = LLaMEA(llm, budget=budget, name=f"LLaMEA-1", mutation_prompts=mutation_prompts1, n_parents=4, n_offspring=12, elitism=False) 
    LLaMEA_method2 = LLaMEA(llm, budget=budget, name=f"LLaMEA-2", mutation_prompts=mutation_prompts2, n_parents=4, n_offspring=12, elitism=False) 
    LLaMEA_method3 = LLaMEA(llm, budget=budget, name=f"LLaMEA-3", mutation_prompts=mutation_prompts3, n_parents=4, n_offspring=12, elitism=False) 
    LLaMEA_method4 = LLaMEA(llm, budget=budget, name=f"LLaMEA-4", mutation_prompts=mutation_prompts3, n_parents=1, n_offspring=1, elitism=True) 
    LLaMEA_method5 = LLaMEA(llm, budget=budget, name=f"LLaMEA-5", mutation_prompts=None, adaptive_mutation=True, n_parents=4, n_offspring=12, elitism=False) 
    LLaMEA_method6 = LLaMEA(llm, budget=budget, name=f"LLaMEA-6", mutation_prompts=None, adaptive_mutation=True, n_parents=1, n_offspring=1, elitism=True) 
# """

# | **Metric**                              | **What it Measures**                                    | **Reference**                          |
# | --------------------------------------- | ------------------------------------------------------- | -------------------------------------- |
# | **Nearest-Neighbor Distance** (avg NND) | Novelty of new samples (exploration)                    | Distance-based exploration     https://arxiv.org/html/2410.14573#:~:text=removing%20points%20from%20the%20Pareto,various%20optimization%20scenarios%20and%20objectives        |
# | **Coverage / Dispersion**               | Fraction of space covered (exploration)                 | Space-filling designs in DOE   https://www.nature.com/articles/s41598-024-68436-1#:~:text=To%20perform%20population%20diversity%20analysis%2C,used%20for%20population%20diversity%20estimation        |
# | **Spatial Entropy**                     | Spread/uncertainty of sample distribution               | Entropy for batch diversity    https://www.nature.com/articles/s41598-024-68436-1#:~:text=To%20perform%20population%20diversity%20analysis%2C,used%20for%20population%20diversity%20estimation        |
# | **Proximity to Best**                   | Focus on known optima (exploitation)                    | Exploitation = staying in neighborhood https://romisatriawahono.net/lecture/rm/survey/softcomputing/Crepinsek%20-%20Exploration%20and%20Exploitation%20in%20Evolutionary%20Algorithms%20-%202013.pdf#:~:text=space,regions%20of%20a%20search%20space  |
# | **Exploration–Exploitation %**          | Trade-off via diversity (time-varying)                  | Pop. diversity ratio                 https://www.nature.com/articles/s41598-024-68436-1#:~:text=%24%24exploration%5C%3B%5Cleft%28iteration%5Cright%29%3D%5Cleft%28%5Cfrac |
# | **Success/Improvement Rate**            | Frequency of improving moves (convergence/exploitation) | 1/5th success rule concept           https://www.nature.com/articles/s41598-024-68436-1#:~:text=%24%24exploration%5C%3B%5Cleft%28iteration%5Cright%29%3D%5Cleft%28%5Cfrac  |
# | **Area Under Conv. Curve**              | Overall speed of convergence                            | – (common in benchmarking)             |
# | **Average Convergence Rate** (ACR)      | Geometric mean error reduction (convergence speed)      | Convergence rate analysis             https://www.nature.com/articles/s41598-024-68436-1#:~:text=%24%24exploration%5C%3B%5Cleft%28iteration%5Cright%29%3D%5Cleft%28%5Cfrac |
# | **No-Improvement Streak**               | Longest stagnation period                               | Used as stopping criterion            https://www.sciencedirect.com/science/article/pii/S037722172200159X#:~:text=The%20second%20parameter%20to%20be,In%20this%20case%2C%20we |
# | **Last Improvement Fraction**           | Portion of run spent stagnating                         | – (derived from convergence curve)     |
# | **Pop. Diversity (to centroid)**        | Dispersion of points in search space                    | “Distance to average” measure         https://www.sciencedirect.com/science/article/pii/S037722172200159X#:~:text=The%20second%20parameter%20to%20be,In%20this%20case%2C%20we |
# | **Mean Pairwise Distance**              | Overall pairwise diversity                              | Diversity measures in EAs             https://romisatriawahono.net/lecture/rm/survey/softcomputing/Crepinsek%20-%20Exploration%20and%20Exploitation%20in%20Evolutionary%20Algorithms%20-%202013.pdf#:~:text=match%20at%20L885%20%E2%80%94Distance,used%20type%20of%20diversity%20measure |
# | **Coverage Volume**                     | Volume spanned by samples (diversity)                   | DPP-based diversity                   https://romisatriawahono.net/lecture/rm/survey/softcomputing/Crepinsek%20-%20Exploration%20and%20Exploitation%20in%20Evolutionary%20Algorithms%20-%202013.pdf#:~:text=match%20at%20L885%20%E2%80%94Distance,used%20type%20of%20diversity%20measure |
# | **Average Step Length**                 | Typical move size (exploration intensity)               | Step size reduction in search         https://stats.stackexchange.com/questions/304813/stopping-criterion-for-nelder-mead#:~:text=Stopping%20criterion%20for%20Nelder%20Mead,than%20some%20tolerance%20TOL |
# | **Total Path Length**                   | Total distance traveled                                 | – (trajectory length analysis)         |
# | **Path Efficiency**                     | Directness of search path                               | –                                      |
# | **Step Length Trend**                   | Change of step size over time (convergence indicator)   | – (implied by 1/5th rule)              |


#df.to_pickle("./BBOB0.pkl")

llm_per_folder = {"BBOB0": "gpt-4.1",
                  "BBOB1": "gemini-2.0-flash",
                  "BBOB2": "o4-mini",
                  "BBOB3": "gemma3-27b"}

for folder in tqdm.tqdm(["BBOB0", "BBOB1", "BBOB2", "BBOB3"]): # 
    #print(f"Processing {folder}...")

    llm_name = llm_per_folder[folder]

    df = pd.read_pickle(f"{folder}.pkl")  
    img_dir = f"examples/img/"
    #df.describe()

    # Some columns you’ll want handy later
    behaviour_feats = [
        "avg_nearest_neighbor_distance", "dispersion", "avg_exploration_pct",
        "avg_distance_to_best", 
        "average_convergence_rate", "avg_improvement", "success_rate",
        "longest_no_improvement_streak", 
    ]
    all_behaviour_feats = [
        "avg_nearest_neighbor_distance", "dispersion", "avg_exploration_pct",
        "avg_distance_to_best",  "intensification_ratio", "avg_exploitation_pct", "last_improvement_fraction",
        "average_convergence_rate", "avg_improvement", "success_rate",
        "longest_no_improvement_streak", 
    ]
    #leave out: 
    agg_map  = {c: 'mean' for c in all_behaviour_feats} | {           # mean for numbers
                'id': 'first',                              # keep id constant
                'fitness_fid': 'mean',                     # average fitness per id
                'fid': 'first',                             
                '_id': 'first',
                'seed': 'first',
                'method_name': 'first',                    # keep one value
                'problem_name': 'first'}                   #  – they should be constant per id
    print(df['_id'].describe())

    df_all = (df
            .groupby('id', as_index=False)
            .agg(agg_map))

    print(df_all['_id'].head())



    # ------------------------------------------------------------------
    # 0  House‑keeping
    # ------------------------------------------------------------------
    # Make a copy we can mangle
    data = df_all.copy()

    # Replace ±∞ with NaN, then drop rows that have no finite fitness at all
    data.replace([np.inf, -np.inf, np.nan], 0, inplace=True)
    #data = data.dropna(subset=["fitness"])

    # Normalise fitness to [0, 1] per algorithm so colour scales are comparable
    data["fitness_norm"] = (
        data.groupby("method_name")["fitness_fid"]
            .transform(lambda s: (s - s.min()) / (s.max() - s.min()))
    )

    
    nice_names = {
        "avg_nearest_neighbor_distance":  "NN-dist",
        "dispersion":                     "Disp",
        "avg_exploration_pct":            "Expl %",
        "avg_distance_to_best":           "Dist→best",
        "intensification_ratio":          "Inten-ratio",
        "avg_exploitation_pct":           "Explt %",
        "average_convergence_rate":       "Conv-rate",
        "avg_improvement":                "Δ fitness",
        "success_rate":                   "Success %",
        "longest_no_improvement_streak":  "No-imp streak",
        "last_improvement_fraction":      "Last-imp frac",
    }

    log_folder = {"BBOB0": "BBOB",
                "BBOB1": "BBOB-1",
                "BBOB2": "BBOB-2",
                "BBOB3": "BBOB-3"}
    loc = log_folder[folder]
    logger = ExperimentLogger(f'/data/neocortex/{loc}', True)
    df_log = logger.get_problem_data(problem_name='BBOB')
    

    data = data.merge(df_log[['id', 'parent_ids']], on='id', how='left') #add parent_ids


    if True:
        # ── NEW BLOCK — STN-style export (drop this *after* `behaviour_feats` is defined
        #    and `data` has been assembled for the current folder) ──────────────────────

        stn_dir = "stn/"
        os.makedirs(stn_dir, exist_ok=True)

        feature_cols = [
            "avg_exploration_pct",
            "average_convergence_rate", "avg_improvement",
            "success_rate",
            "longest_no_improvement_streak",
        ]  # the columns that go into the STN feature list (subset)
        #         "avg_nearest_neighbor_distance", "dispersion", "avg_exploration_pct",
        # "avg_distance_to_best",  "intensification_ratio", "avg_exploitation_pct", "last_improvement_fraction",
        # "average_convergence_rate", "avg_improvement", "success_rate",
        # "longest_no_improvement_streak", 

        data_stn = data.copy()
        data_stn[behaviour_feats] = robust_minmax(data_stn, behaviour_feats)
        for method, g in data_stn.groupby("method_name", sort=False):


            lines = []
            lines_best = []
            # one file line per evaluation, ordered by run (seed) then _id
            for seed, sg in g.groupby("seed", sort=False):
                # Quick lookup table: id  ➜ row (Series)
                #id_lookup = sg.set_index("id", drop=False)
                id_lookup = (
                    sg.sort_values("_id")           # ensures _id ascending
                    .drop_duplicates("id", keep="first")
                    .set_index("id", drop=False)
                )

                current_best = -np.inf
                for _, child in sg.sort_values("_id").iterrows():
                    # print(child)
                    # a =aaaa

                    # ── get parent row, if any ───────────────────────────────
                    parent = None
                    parent_list = child["parent_ids"]
                    if isinstance(parent_list, (list, tuple)) and parent_list:
                        pid = parent_list[0]
                        if pid in id_lookup.index:
                            parent = id_lookup.loc[pid]
                    
                    if parent is None:
                        # No parent found, use the current child as parent
                        parent = child
                            
                    # ── stringify parent & child info ───────────────────────
                    # parent block
                    if parent is not None:
                        try:
                            feat_str_parent = ",".join(f"{parent[c]:.12g}" for c in feature_cols)
                            parent_block = f"{parent['fitness_fid']:.12g}\t{feat_str_parent}"
                            # child block
                            feat_str_child = ",".join(f"{child[c]:.12g}" for c in feature_cols)
                            child_block = f"{child['fitness_fid']:.12g}\t{feat_str_child}"
                            # complete line: seed ⟶ parent ⟶ child
                            lines.append(f"{seed}\t{parent_block}\t{child_block}")
                        except Exception as e:
                            print(f"Error processing seed {seed} with parent {pid}: {e}")
                            print(parent)
                            continue

                    # feat_str = ",".join(f"{row[c]:.12g}" for c in feature_cols)
                    # lines.append(f"{seed}\t{row['fitness_fid']:.12g}\t{feat_str}")
                    # if row["fitness_fid"] > current_best:
                    #     current_best = row["fitness_fid"]
                    #     lines_best.append(f"{seed}\t{row['fitness_fid']:.12g}\t{feat_str}")

            fname = f"{llm_name}_{method}_all.csv".replace(" ", "_")
            with open(os.path.join(stn_dir, fname), "w") as fh:
                fh.write("\n".join(lines))

        #continue #only do STN for now

            # fname_best = f"{llm_name}_{method}_best.csv".replace(" ", "_")
            # with open(os.path.join(stn_dir, fname_best), "w") as fh:
            #     fh.write("\n".join(lines_best))



        # ------------------------------------------------------------------
        # 0  Pre-filter: keep only rows where fitness_fid strictly improves
        #    relative to the previous evaluation of the *same seed*.
        # ------------------------------------------------------------------
        improving = (
            data.sort_values(["seed", "_id"])          # just to be safe
                .groupby("seed")["fitness_fid"]
                .apply(lambda s: s > s.shift())        # True where improvement
                .reset_index(level=0, drop=True)       # align with original index
        )

        data_imp = data[improving].copy()              # <- only improving evals

            
        # ------------------------------------------------------------------
        # 1  Visualise behaviour features *along those improving steps only*
        # ------------------------------------------------------------------
        for algo, g in data_imp.groupby("method_name", sort=False):
            g = g.sort_values(["seed", "_id"])          # nicer lines

            for feat in behaviour_feats:
                plt.figure(figsize=(8, 4))

                # individual improving stretches
                for _, seed_group in g.groupby("seed"):
                    plt.plot(
                        seed_group["_id"],
                        seed_group[feat],
                        linewidth=0.6,
                        alpha=0.35,
                    )

                # mean trend of improving steps
                mean_curve = (
                    g.groupby("_id")[feat]
                    .mean()
                    .reset_index()
                    .sort_values("_id")
                )
                plt.plot(
                    mean_curve["_id"],
                    mean_curve[feat],
                    linewidth=2,
                    label="mean on improving evals",
                )

                plt.title(f"{algo}")
                plt.xlabel("Evaluations")
                plt.ylabel(f"{feat} - of best so far")
                plt.legend()
                plt.tight_layout()
                plt.savefig(f"{img_dir}{llm_name}_{algo}_{feat}_improving.png")
                plt.clf()
                plt.close()
        
        # ------------------------------------------------------------------
        # 1  Behaviour features over evaluations
        #     (_id is assumed to be monotonically increasing per seed)
        # ------------------------------------------------------------------
        for algo, g in data.groupby("method_name", sort=False):
            # Pre‑sort once for nicer lines
            g = g.sort_values(["seed", "_id"])

            for feat in behaviour_feats:
                plt.figure(figsize=(8, 4))

                # All individual runs in the background
                for _, seed_group in g.groupby("seed"):
                    plt.plot(
                        seed_group["_id"],
                        seed_group[feat],
                        linewidth=0.6,
                        alpha=0.3,
                    )

                # Mean trend over runs on top
                mean_curve = (
                    g.groupby("_id")[feat]
                    .mean()
                    .reset_index()
                    .sort_values("_id")
                )
                plt.plot(
                    mean_curve["_id"],
                    mean_curve[feat],
                    linewidth=2,
                    label="per‑eval mean",
                )

                plt.title(f"{algo}")
                plt.xlabel("Evaluations")
                plt.ylabel(feat)
                plt.legend(loc="best")
                plt.tight_layout()
                plt.savefig(f"{img_dir}{llm_name}_{algo}_{feat}.png")
                plt.clf()
                plt.close()

        # ------------------------------------------------------------------
        # 2  All methods in one figure per behaviour feature
        # ------------------------------------------------------------------
        palette = sns.color_palette("tab10", n_colors=data["method_name"].nunique())

        for feat in behaviour_feats:
            plt.figure(figsize=(10, 6))

            # Average each method over seeds at every evaluation step
            mean_per_method = (
                data.groupby(["method_name", "_id"])[feat]
                    .median()
                    .reset_index()
                    .sort_values("_id")
            )

            # One line per algorithm
            for (algo, g), colour in zip(mean_per_method.groupby("method_name", sort=False),
                                        palette):
                plt.plot(
                    g["_id"],
                    g[feat],
                    label=algo,
                    linewidth=2,
                    alpha=0.8,
                    color=colour,
                )

            plt.title(f"Median {nice_names.get(feat, feat)}")
            plt.xlabel("Evaluations")
            plt.ylabel(nice_names.get(feat, feat))
            plt.legend(loc="best", frameon=False, fontsize="small")
            plt.tight_layout()
            plt.savefig(f"{img_dir}{llm_name}_ALL_{feat}.png")
            plt.clf()
            plt.close()

    if False:
        # ------------------------------------------------------------------
        # 1  Correlation heat‑map 
        # ------------------------------------------------------------------
        #for algo, g in data.groupby("method_name", sort=False):

        nice_names_corr = {
            "avg_nearest_neighbor_distance":  "NN-dist",
            "dispersion":                     "Disp",
            "avg_exploration_pct":            "Expl %",
            "avg_distance_to_best":           "Dist→best",
            "intensification_ratio":          "Inten-ratio",
            "avg_exploitation_pct":           "Explt %",
            "average_convergence_rate":       "Conv-rate",
            "avg_improvement":                "Δ fitness",
            "success_rate":                   "Success %",
            "longest_no_improvement_streak":  "No-imp streak",
            "last_improvement_fraction":      "Last-imp frac",
            "fitness_norm":                   "Normalized Fitness",
        }

        
        corr_data = data.copy()
        corr_data[all_behaviour_feats] = robust_minmax(corr_data, all_behaviour_feats)
        corr_data.rename(columns=nice_names_corr, inplace=True)
        
        corr_data.drop(columns=["id", "fitness_fid", "fid", "_id", "seed", "method_name", "problem_name", 'parent_ids'], inplace=True)

        corr_data.dropna(inplace=True)  # drop rows with NaN values
        plot_df = corr_data.corr()

        plt.figure(figsize=(16, 14))
        sns.heatmap(
            plot_df,
            cmap="coolwarm",
            annot=True,
            fmt=".1f",
            center=0,
            vmin=-1,
            vmax=1,
            square=True,
            linewidths=0.5,
            cbar_kws=dict(label="Pearson r"),
        )
        plt.title(f"Behaviour metrics\ncorrelation matrix")
        plt.tight_layout()
        plt.savefig(f"{img_dir}{llm_name}_behaviour-metrics-correlation.png")
        #plt.show()
        plt.clf()
    
    if False:
        # ------------------------------


        # Parallel coordinates for all!
        pc = data[behaviour_feats + ["fitness_norm"]].copy()

        # 2‑a.  Min‑max scale features (parallel plots hate disparate ranges)
        # pc[behaviour_feats] = (
        #     pc[behaviour_feats] - pc[behaviour_feats].min()
        # ) / (pc[behaviour_feats].max() - pc[behaviour_feats].min())
        pc[behaviour_feats] = robust_minmax(pc, behaviour_feats)

        quartile_labels = ["Q1 (low)", "Q2", "Q3", "Q4 (high)"]
        quart_cat = pd.CategoricalDtype(categories=quartile_labels, ordered=True)

        # 2‑b.  Quartile‑bin fitness so we can colour by performance
        pc["fitness_group"] = pd.qcut(
            pc["fitness_norm"], 4, labels=quartile_labels, duplicates="drop"
        )
        pc["fitness_group"] = pc["fitness_group"].astype(quart_cat)
        pc = pc.sort_values("fitness_group", key=lambda s: s.cat.codes)

        # ── optional: subsample to keep the plot readable ──────────────
        #sample = pc.sample(n=min(len(pc), 600), random_state=42)

        fig, ax = plt.subplots(figsize=(16, 12))

        plot_df = pc.rename(columns=nice_names)
        parallel_coordinates(
            plot_df, "fitness_group", alpha=0.3, linewidth=1.0,
            colormap="seismic", ax=ax,
        )

        handles, _ = ax.get_legend_handles_labels()
        # ax.legend(handles, quartile_labels, title="fitness quartile",
        #     loc="upper left", frameon=True)
        plt.title(f"Behaviour profile")
        plt.ylabel("scaled feature value")
        plt.xticks(rotation=90, ha="right")
        plt.tight_layout()
        plt.savefig(f"{img_dir}{llm_name}_behaviour-pc.png")
        plt.clf()
        plt.close()

        # ------------------------------------------------------------------
        # 2  Parallel‑coordinate plot per algorithm
        # ------------------------------------------------------------------
        for algo, g in data.groupby("method_name", sort=False):

            pc = g[behaviour_feats + ["fitness_norm"]].copy()

            # 2‑a.  Min‑max scale features (parallel plots hate disparate ranges)
            # pc[behaviour_feats] = (
            #     pc[behaviour_feats] - pc[behaviour_feats].min()
            # ) / (pc[behaviour_feats].max() - pc[behaviour_feats].min())
            pc[behaviour_feats] = robust_minmax(pc, behaviour_feats)

            # 2‑b.  Quartile‑bin fitness so we can colour by performance
            pc["fitness_group"] = pd.qcut(
                pc["fitness_norm"], 4, labels=quartile_labels, duplicates="drop" #["Q1 (low)", "Q2", "Q3", "Q4 (high)"]
            )
            pc["fitness_group"] = pc["fitness_group"].astype(quart_cat)

            pc = pc.sort_values("fitness_group", key=lambda s: s.cat.codes)
            # ── optional: subsample to keep the plot readable ──────────────
            #sample = pc.sample(n=min(len(pc), 600), random_state=42)

            plt.figure(figsize=(16, 12))
            parallel_coordinates(
                pc, "fitness_group", alpha=0.4, linewidth=1.0,
                colormap="seismic",
            )
            plt.title(f"{algo} — parallel‑coordinate behaviour profile")
            plt.ylabel("scaled feature value")
            plt.xticks(rotation=90, ha="right")
            plt.tight_layout()
            plt.savefig(f"{img_dir}{llm_name}_{algo}_pc.png")
            plt.clf()
            plt.close()

        # ------------------------------------------------------------------
        # 0  Aggregate the 50 raw auc‑columns into 10 function‑level scores
        # ------------------------------------------------------------------



        func_ids        = [1, 3, 6, 8, 10, 13, 15, 17, 21, 23]
        runs_per_func   = 5          # we know each function has 5 runs
        quart_labels    = ["Q1 (low)", "Q2", "Q3", "Q4 (high)"]
        quart_cat       = pd.CategoricalDtype(quart_labels, ordered=True)



        # ------------------------------------------------------------------
        # 1  Build a helper with min‑max‑scaled behaviour features
        # ------------------------------------------------------------------
        behaviour_feats = [
            "avg_nearest_neighbor_distance", "dispersion", "avg_exploration_pct",
            "avg_distance_to_best", "intensification_ratio", "avg_exploitation_pct",
            "average_convergence_rate", "avg_improvement", "success_rate",
            "longest_no_improvement_streak", "last_improvement_fraction",
        ]

        nice_names = {
            "avg_nearest_neighbor_distance":  "NN-dist",
            "dispersion":                     "Disp",
            "avg_exploration_pct":            "Expl %",
            "avg_distance_to_best":           "Dist→best",
            "intensification_ratio":          "Inten-ratio",
            "avg_exploitation_pct":           "Explt %",
            "average_convergence_rate":       "Conv-rate",
            "avg_improvement":                "Δ fitness",
            "success_rate":                   "Success %",
            "longest_no_improvement_streak":  "No-imp streak",
            "last_improvement_fraction":      "Last-imp frac",
        }

        data = df.copy()
        # data[behaviour_feats] = (
        #     data[behaviour_feats] - data[behaviour_feats].min()
        # ) / (data[behaviour_feats].max() - data[behaviour_feats].min())
        data[behaviour_feats] = robust_minmax(data, behaviour_feats)

        # ------------------------------------------------------------------
        # 2  Parallel‑coordinate plot for each function‑level AUC
        # ------------------------------------------------------------------
        from pandas.plotting import parallel_coordinates
        import matplotlib.pyplot as plt

        col = "fitness_fid"
        for fid in func_ids:
            num_cols = data.select_dtypes('number').columns          # all numeric columns
            agg_map  = {c: 'mean' for c in num_cols} | {           # mean for numbers
                        'method_name': 'first',                    # keep one value
                        'problem_name': 'first'}                   #  – they should be constant per id

            df_fid = data[data["fid"] == fid].copy()
            df_f = (df_fid
                    .groupby('id', as_index=False)
                    .agg(agg_map))


            if df_f.empty:       # nothing to draw for this function
                continue



            # quartile‑bin the aggregated fitness score
            q = pd.qcut(df_f[col], 4, labels=quart_labels).astype(quart_cat)

            # frame to plot: scaled behaviour + quartile label
            plot_df = df_f[behaviour_feats + ["fitness_fid"]].copy()
            plot_df["fitness_quartile"] = q

            # (optional) thin out lines for readability
            #plot_df = plot_df.sample(n=min(len(plot_df), 2000), random_state=42)
            plot_df = plot_df.sort_values("fitness_quartile", key=lambda s: s.cat.codes)

            fig, ax = plt.subplots(figsize=(16, 12))
            
            plot_df = plot_df.rename(columns=nice_names)
            parallel_coordinates(
                plot_df, "fitness_quartile", ax=ax,
                alpha=0.35, linewidth=1.0, colormap="seismic",
            )

            # handles, _ = ax.get_legend_handles_labels()
            # ax.legend(handles, quart_labels, title="fitness quartile",
            #           loc="upper left", frameon=True)

            ax.set_title(f"Behaviour profile for $f_{{{fid}}}$")
            ax.set_ylabel("scaled feature value")
            ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha="right")
            plt.tight_layout()
            plt.savefig(f"{img_dir}{llm_name}_f{fid}_pc.png")
            plt.clf()
            plt.close()

        TOP_K     = 100            # ← change this to whatever "top‑k" you want

        # ---------------------------------------------------------------
        # 1  Gather top‑k rows per function (by aggregated auc_f*)
        # ---------------------------------------------------------------
        pieces = []

        behaviour_feats = [
            "avg_nearest_neighbor_distance", "dispersion", "avg_exploration_pct",
            "avg_distance_to_best", "intensification_ratio", "avg_exploitation_pct",
            #"average_convergence_rate", "avg_improvement", "success_rate",
            "longest_no_improvement_streak", "last_improvement_fraction",
        ]
        nice_names = {
            "avg_nearest_neighbor_distance":  "NN-dist",
            "dispersion":                     "Disp",
            "avg_exploration_pct":            "Expl %",
            "avg_distance_to_best":           "Dist→best",
            "intensification_ratio":          "Inten-ratio",
            "avg_exploitation_pct":           "Explt %",
        #    "average_convergence_rate":       "Conv-rate",
        #    "avg_improvement":                "Δ fitness",
        #    "success_rate":                   "Success %",
            "longest_no_improvement_streak":  "No-imp streak",
            "last_improvement_fraction":      "Last-imp frac",
        }

        col = "fitness_fid"
        for fid in func_ids:

            df_fid = data[data["fid"] == fid].copy()
            df_f = (df_fid
                    .groupby('id', as_index=False)
                    .agg(agg_map))

            # take the k rows with highest score for this function
            topk = df_f.nlargest(TOP_K, columns=col)

            # scaled behaviour + fid label
            
            frame = df_f[behaviour_feats + ["fitness_fid"]].loc[topk.index].copy()
            frame["fid"] = f"f{fid}"
            pieces.append(frame)

        # Concatenate every function’s Q4 slice
        plot_df = pd.concat(pieces, ignore_index=True)

        # Rename cols for prettier x‑ticks
        plot_df = plot_df.rename(columns=nice_names)

        # ------------------------------------------------------------------
        # 3  Parallel‑coordinate plot
        # ------------------------------------------------------------------
        from pandas.plotting import parallel_coordinates
        import matplotlib.pyplot as plt

        # Build a colour list (one per fid) – tab10 has 10 distinct hues
        cmap   = plt.get_cmap("tab10")
        colors = [cmap(i) for i in range(len(func_ids))]

        fig, ax = plt.subplots(figsize=(16, 12))
        parallel_coordinates(
            plot_df, "fid",
            ax=ax,
            linewidth=1.1,
            alpha=0.9,
            color=colors,
        )

        ax.set_title("Top-100 behaviour profiles per function id")
        ax.set_ylabel("scaled feature value")
        ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha="right")
        ax.legend(title="function id", frameon=True, loc="upper left")
        plt.tight_layout()
        plt.savefig(f"{img_dir}{llm_name}_fids_pc.png")
        plt.clf()
        plt.close()

        # BLADE loggers
        log_folder = {"BBOB0": "BBOB",
                    "BBOB1": "BBOB-1",
                    "BBOB2": "BBOB-2",
                    "BBOB3": "BBOB-3"}
        loc = log_folder[folder]
        logger = ExperimentLogger(f'/data/neocortex/{loc}', True)
        plot_convergence(logger, metric="AOCC", save=True, budget=100)
        plot_experiment_CEG(logger, save=True, budget=100, max_seeds=5)




import json
import math
import os
import random
import re
import time
import traceback

import numpy as np
import pandas as pd
import polars as pl
import sklearn
from sklearn.datasets import load_breast_cancer
from sklearn.metrics import accuracy_score

from ..problem import Problem
from ..solution import Solution

# import autosklearn.classification


class AutoML(Problem):
    """
    Problem class for evaluating AutoML pipelines (sample).

    """

    def __init__(
        self, logger=None, datasets=None, name="AutoML-breast_cancer", eval_timeout=360
    ):
        X, y = load_breast_cancer(return_X_y=True)
        (
            self.X_train,
            self.X_test,
            self.y_train,
            self.y_test,
        ) = sklearn.model_selection.train_test_split(X, y, random_state=1)

        super().__init__(
            logger,
            [(self.X_train, self.y_train)],
            [(self.X_test, self.y_test)],
            name,
            eval_timeout,
        )
        self.func_name = "__call__"
        self.init_inputs = ["X", "y"]
        self.func_inputs = ["X"]
        self.func_outputs = ["y_pred"]

        self.task_prompt = f"""
You are a highly skilled computer scientist in the field machine learning. Your task is to design novel machine learning pipelines for a given dataset and task.
The pipeline in this case should handle a breast cancer classification task. Your task is to write the Python code. The code should contain an `__init__(self, X, y)` function that trains a machine learning model and the function `def __call__(self, X)`, which should predict the samples in X and return the predictions.
The training data X has shape {self.X_train.shape} and y has shape {self.y_train.shape}.
"""
        self.example_prompt = """
An example code structure is as follows:
```python
import numpy as np
import sklearn

class AlgorithmName:
    "Template for a ML pipeline"

    def __init__(self, X, y):
        self.train(X, y)

    def train(self, X, y):
        # Standardize the feature data
        scaler = sklearn.preprocessing.StandardScaler()
        X_train = scaler.fit_transform(X)

        # Let's create and train a logistic regression model
        lr_model = sklearn.linear_model.LogisticRegression()
        lr_model.fit(X_train, y)
        self.model = lr_model
        
    def __call__(self, X):
        # predict using the trained model
        return self.model.predict(X)
```
"""
        self.format_prompt = """

Give an excellent and novel ML pipeline to solve this task and also give it a one-line description, describing the main idea. Give the response in the format:
# Description: <short-description>
# Code: 
```python
<code>
```
"""

    def get_prompt(self):
        """
        Returns the problem description and answer format.
        """
        return self.task_prompt + self.example_prompt + self.format_prompt

    def evaluate(self, solution: Solution, test=False, ioh_dir=""):
        """
        Evaluates a solution on the kernel tuner benchmark using AOCC.
        """
        code = solution.code
        algorithm_name = solution.name
        safe_globals = {
            "sklearn": sklearn,
            "math": math,
            "random": random,
            "np": np,
            "pd": pd,
        }

        exec(code, globals())

        algorithm = None

        # Final validation
        algorithm = globals()[algorithm_name](self.X_train, self.y_train)
        y_pred = algorithm(self.X_test)
        score = accuracy_score(self.y_test, y_pred)

        solution.set_scores(
            score,
            f"The algorithm {algorithm_name} scored {score:.3f} on accuracy (higher is better, 1.0 is the best).",
        )

        return solution

    def test(self, solution: Solution, ioh_dir=""):
        """
        Runs the solution on test instances and returns the fitness score.
        """
        return self.evaluate(solution, True, ioh_dir)

    def to_dict(self):
        """
        Converts the problem to a dictionary.
        """
        return {
            "name": self.name,
        }



import json
import os
from datetime import datetime

import wandb
from ConfigSpace.read_and_write import json as cs_json

from ..llm import LLM
from ..method import Method
from ..problem import Problem
from ..solution import Solution
from ..utils import convert_to_serializable
from .base import ExperimentLogger, RunLogger


class WAndBExperimentLogger(ExperimentLogger):
    """
    An ExperimentLogger subclass that *also* logs to Weights & Biases (W&B).
    The flow is:
      1. Call open_run() to start a W&B run (wandb.init()).
      2. Perform your optimization (the RunLogger logs intermediate data).
      3. Call add_run() to log final info and finish the W&B run,
         plus the original file-based logging.
    """

    def __init__(self, name="", entity=None, project=None, read=False):
        """
        Args:
            name (str): The name of the experiment. Used in W&B as the 'group' or 'project' name if you prefer.
            entity (str): W&B username or team where you're sending runs.
            project (str): W&B project name.
            read (bool): If True, read logs from an existing directory (original file-based usage).
        """
        # Still call the parent constructor so we keep file-based logging
        super().__init__(name=name, read=read)
        self.wandb_entity = entity
        self.wandb_project = project if project else "DefaultProject"
        self._wandb_run_active = False

    def open_run(self, method, problem, budget=100, seed=0):
        """
        Opens (starts) a new MLflow run for logging.
        Typically call this right before your run, so that the RunLogger can log step data.
        """
        if self._wandb_run_active:
            print("Warning: An MLflow run is already active. Not starting a new one.")
            return
        run_name = f"{method.name}-{problem.name}-{seed}"

        wandb.init(
            entity=self.wandb_entity,
            project=self.wandb_project,
            name=run_name,
            config={},
            # group=name to group runs in your W&B dashboard, if desired
        )
        self._wandb_run_active = True

        self.run_logger = WAndBRunLogger(
            name=run_name,
            root_dir=self.dirname,
            budget=budget,
            progress_callback=lambda: self.increment_eval(
                method.name, problem.name, seed
            ),
        )
        problem.set_logger(self.run_logger)
        return self.run_logger

    def add_run(
        self,
        method: Method,
        problem: Problem,
        llm: LLM,
        solution: Solution,
        log_dir="",
        seed=None,
    ):
        """
        Called at the end of a run:
          - Logs final data to W&B (if active).
          - Finishes (wandb.finish()) the run.
          - Calls super().add_run(...) to keep file-based logs.
        """
        # If no run is active, open one automatically
        if not self._wandb_run_active:
            self.open_run(method, problem, seed=seed)

        # Log final parameters
        wandb.config.update({"method_name": method.name}, allow_val_change=True)
        wandb.config.update({"problem_name": problem.name}, allow_val_change=True)
        wandb.config.update({"llm_name": llm.model}, allow_val_change=True)
        if seed is not None:
            wandb.config.update({"seed": seed}, allow_val_change=True)

        # Log final fitness as a metric
        final_fitness = (
            solution.fitness if solution.fitness is not None else float("nan")
        )
        wandb.log({"final_fitness": final_fitness})

        # Log a serialized run object
        rel_log_dir = os.path.relpath(log_dir, self.dirname)
        run_object = {
            "method_name": method.name,
            "problem_name": problem.name,
            "llm_name": llm.model,
            "method": method.to_dict(),
            "problem": problem.to_dict(),
            "llm": llm.to_dict(),
            "solution": solution.to_dict(),
            "log_dir": rel_log_dir,
            "seed": seed,
        }
        # For large data, consider W&B artifacts. For quick usage, we do a single JSON log:
        wandb.log({"final_run_object": run_object})

        # End the W&B run
        wandb.finish()
        self._wandb_run_active = False

        # Also do file-based logging
        super().add_run(
            method=method,
            problem=problem,
            llm=llm,
            solution=solution,
            log_dir=log_dir,
            seed=seed,
        )


class WAndBRunLogger(RunLogger):
    """
    A RunLogger subclass that logs intermediate data to Weights & Biases while
    preserving the original file-based logs.
    """

    def __init__(self, name="", root_dir="", budget=100, progress_callback=None):
        """
        As before, call super() so we keep the local directories.
        """
        super().__init__(
            name=name,
            root_dir=root_dir,
            budget=budget,
            progress_callback=progress_callback,
        )

    def log_conversation(self, role, content, cost=0.0, tokens=0):
        """
        Log conversation data to W&B plus the normal local file-based logs.
        """
        conversation = {
            "role": role,
            "time": str(datetime.now()),
            "content": content,
            "cost": float(cost),
            "tokens": int(tokens),
        }
        # wandb.log can be repeated, but if content is large, consider an artifact
        wandb.log({"conversation": conversation})

        # Also do file-based logging
        super().log_conversation(role, content, cost, tokens)

    def log_individual(self, individual: Solution):
        """
        Log an individual solution's data/fitness to W&B.
        Then call super() for file-based logging.
        """
        ind_dict = individual.to_dict()

        # Log the fitness as a metric
        if "fitness" in ind_dict:
            wandb.log({"fitness": ind_dict["fitness"]})

        # Also log the entire solution as a dict
        wandb.log({"solution": ind_dict})

        # Continue with file-based logging
        super().log_individual(individual)

    def log_code(self, individual: Solution):
        """
        For code, you can store it as text in W&B. Large code might benefit from an artifact approach.
        """
        code_text = individual.code
        # For short code, we can do:
        wandb.log({f"code_{individual.id}": code_text})
        # If it's large or you want versioning, you'd do a W&B artifact instead.

        # File-based approach
        super().log_code(individual)

    def log_configspace(self, individual: Solution):
        """
        Log configspace as text in W&B, plus file-based logging.
        """
        if individual.configspace is not None:
            cs_text = cs_json.write(individual.configspace)
        else:
            cs_text = "Failed to extract config space"
        wandb.log({f"configspace_{individual.id}": cs_text})

        super().log_configspace(individual)

    def budget_exhausted(self):
        """
        Optionally keep using the parent's file-based line counting logic, or switch to an in-memory count.
        """
        return super().budget_exhausted()



import os
import sys

import ioh
import numpy as np

from .problems.brag_mirror import brag_mirror
from .problems.ellipsometry_inverse import ellipsometry
from .problems.grating2D import grating2D
from .problems.plasmonic_nanostructure import plasmonic_nanostructure
from .problems.sophisticated_antireflection_design import (
    sophisticated_antireflection_design,
)

problem_descriptions = {
    "bragg": "The Bragg mirror optimization aims to maximize reflectivity at a wavelength of 600 nm using a multilayer structure with alternating refractive indices (1.4 and 1.8). The structure's thicknesses are varied to find the configuration with the highest reflectivity. The problem involves two cases: one with 10 layers (minibragg) and another with 20 layers (bragg), with the latter representing a more complex inverse design problem. The known optimal solution is a periodic Bragg mirror, which achieves the best reflectivity by leveraging constructive interference. This case exemplifies challenges such as multiple local minima in the optimization landscape. ",
    "ellipsometry": "The ellipsometry problem involves retrieving the material and thickness of a reference layer by matching its reflectance properties using a known spectral response. The optimization minimizes the difference between the calculated and measured ellipsometric parameters for wavelengths between 400 and 800 nm and a fixed incidence angle of 40°. The parameters to be optimized include the thickness (30 to 250 nm) and refractive index (1.1 to 3) of the test layer. This relatively straightforward problem models a practical scenario where photonics researchers fine-tune a small number of parameters to achieve a desired spectral fit. ",
    "photovoltaic": "The photovoltaics problem optimizes the design of an antireflective multilayer coating to maximize the absorption in the active silicon layer of a solar cell. The goal is to achieve maximum short-circuit current in the 375 to 750 nm wavelength range. The structure consists of alternating materials with permittivities of 2 and 3, built upon a 30,000 nm thick silicon substrate. Three subcases with increasing complexity are explored, involving 10 layers (photovoltaics), 20 layers (bigphotovoltaics), and 32 layers (hugephotovoltaics). The optimization challenges include balancing high absorption with a low reflectance while addressing the inherent noise and irregularities in the solar spectrum. ",
}
algorithmic_insights = {
    "bragg": "For this problem, the optimization landscape contains multiple local minima due to the wave nature of the problem. And periodic solutions are known to provide near-optimal results, suggesting the importance of leveraging constructive interference principles. Here are some suggestions for designing algorithms: 1. Use global optimization algorithms like Differential Evolution (DE) or Genetic Algorithms (GA) to explore the parameter space broadly. 2. Symmetric initialization strategies (e.g., Quasi-Oppositional DE) can improve exploration by evenly sampling the search space. 3. Algorithms should preserve modular characteristics in solutions, as multilayer designs often benefit from distinct functional blocks. 4. Combine global methods with local optimization (e.g., BFGS) to fine-tune solutions near promising regions. 5. Encourage periodicity in solutions via tailored cost functions or constraints. ",
    "ellipsometry": "This problem has small parameter space with fewer variables (thickness and refractive index), and the cost function is smooth and relatively free of noise, making it amenable to local optimization methods. Here are suggestions for designing algorithms: 1. Use local optimization algorithms like BFGS or Nelder-Mead, as they perform well in low-dimensional, smooth landscapes. 2. Uniform sampling across the parameter space ensures sufficient coverage for initial guesses. 3. Utilize fast convergence algorithms that can quickly exploit the smooth cost function landscape. 4. Iteratively adjust bounds and constraints to improve parameter estimates once initial solutions are obtained. ",
    "photovoltaic": "This problem is a challenging high-dimensional optimization problem with noisy cost functions due to the realistic solar spectrum, and it requires maximizing absorption while addressing trade-offs between reflectance and interference effects. Here are the suggestions for designing algorithms: 1. Combine global methods (e.g., DE, CMA-ES) for exploration with local optimization for refinement. 2. Use consistent benchmarking and convergence analysis to allocate computational resources effectively. 3. Encourage algorithms to detect and preserve modular structures (e.g., layers with specific roles like anti-reflective or coupling layers). 4. Gradually increase the number of layers during optimization to balance problem complexity and computational cost. 5. Integrate robustness metrics into the cost function to ensure the optimized design tolerates small perturbations in layer parameters. ",
}


def get_photonic_instance(problem_name="bragg"):
    if problem_name == "bragg":
        # ------- define "mini-bragg" optimization problem
        nb_layers = 10  # number of layers of full stack
        target_wl = 600.0  # nm
        mat_env = 1.0  # materials: ref. index
        mat1 = 1.4
        mat2 = 1.8
        prob = brag_mirror(nb_layers, target_wl, mat_env, mat1, mat2)
        ioh.problem.wrap_real_problem(
            prob, name="brag_mirror", optimization_type=ioh.OptimizationType.MIN
        )
        problem = ioh.get_problem("brag_mirror", dimension=prob.n)
        problem.bounds.lb = prob.lb
        problem.bounds.ub = prob.ub
    elif problem_name == "ellipsometry":
        # ------- define "ellipsometry" optimization problem
        mat_env = 1.0
        mat_substrate = "Gold"
        nb_layers = 1
        min_thick = 50  # nm
        max_thick = 150
        min_eps = 1.1  # permittivity
        max_eps = 3
        wavelengths = np.linspace(400, 800, 100)  # nm
        angle = 40 * np.pi / 180  # rad
        prob = ellipsometry(
            mat_env,
            mat_substrate,
            nb_layers,
            min_thick,
            max_thick,
            min_eps,
            max_eps,
            wavelengths,
            angle,
        )
        ioh.problem.wrap_real_problem(
            prob,
            name="ellipsometry",
            optimization_type=ioh.OptimizationType.MIN,
        )
        problem = ioh.get_problem("ellipsometry", dimension=prob.n)
        problem.bounds.lb = prob.lb
        problem.bounds.ub = prob.ub
    elif problem_name == "photovoltaic":
        # ------- define "sophisticated antireflection" optimization problem
        nb_layers = 10
        min_thick = 30
        max_thick = 250
        wl_min = 375
        wl_max = 750
        prob = sophisticated_antireflection_design(
            nb_layers, min_thick, max_thick, wl_min, wl_max
        )
        ioh.problem.wrap_real_problem(
            prob,
            name="sophisticated_antireflection_design",
            optimization_type=ioh.OptimizationType.MIN,
        )
        problem = ioh.get_problem(
            "sophisticated_antireflection_design", dimension=prob.n
        )
        problem.bounds.lb = prob.lb
        problem.bounds.ub = prob.ub
    return problem



import numpy as np
import PyMoosh as pm

from .photonic_problem import photonic_problem


class brag_mirror(photonic_problem):
    def __init__(self, nb_layers, target_wl, mat_env, mat1, mat2):
        """
        Args:
            mat_env (float): environment ref. index
            mat1 (float): material 1 ref. index
            mat2 (float): material 2 ref. index
        """
        super().__init__()
        self.n = nb_layers
        self.nb_layers = nb_layers
        self.target_wl = target_wl
        self.mat_env = mat_env
        self.mat1 = mat1
        self.mat2 = mat2
        self.min_thick = 0
        self.max_thick = target_wl / (2 * mat1)
        self.lb = np.array([self.min_thick] * self.n)
        self.ub = np.array([self.max_thick] * self.n)

    def setup_structure(self, x):
        """helper to create pymoosh structure object, alternating 2 materials

        Args:
            x (list): list of thicknesses, top layer first

        Returns:
            PyMoosh.structure: multi-layer structure object
        """
        x = list(x)
        # n = len(x)
        materials = [
            self.mat_env**2,
            self.mat1**2,
            self.mat2**2,
        ]  # permittivities!
        # periodic stack. first layer: environment, last layer: substrate
        stack = [0] + [2, 1] * (self.n // 2) + [2]
        thicknesses = [0.0] + x + [0.0]
        structure = pm.Structure(materials, stack, np.array(thicknesses), verbose=False)
        return structure

    def __call__(self, x):
        """cost function: maximize reflectance of a layer-stack

        Args:
            x (list): thicknesses of all the layers, starting with the upper one.

        Returns:
            float: 1 - Reflectivity at target wavelength
        """
        x = np.clip(x, self.lb, self.ub)
        structure = self.setup_structure(x)
        # the actual PyMoosh reflectivity simulation
        _, R = pm.coefficient_I(structure, self.target_wl, 0.0, 0)
        cost = 1 - R

        return cost



import json
import os
import subprocess
import time
from pathlib import Path

import matplotlib
import pandas as pd
import plotly.graph_objects as go
import streamlit as st

from iohblade.assets import LOGO_DARK_B64, LOGO_LIGHT_B64
from iohblade.loggers import ExperimentLogger

LOGO_LIGHT = f"data:image/png;base64,{LOGO_LIGHT_B64}"
LOGO_DARK = f"data:image/png;base64,{LOGO_DARK_B64}"


def convergence_dataframe(logger: ExperimentLogger) -> pd.DataFrame:
    methods, problems = logger.get_methods_problems()
    frames = []
    for p in problems:
        df = logger.get_problem_data(problem_name=p).drop(columns=["code"])
        df.replace([-pd.NA, float("inf"), float("-inf")], 0, inplace=True)
        df["problem_name"] = p
        frames.append(df)
    if not frames:
        return pd.DataFrame()
    df = pd.concat(frames, ignore_index=True)
    df["cummax_fitness"] = df.groupby(["method_name", "problem_name", "seed"])[
        "fitness"
    ].cummax()
    df["eval"] = df["_id"] + 1
    return df


def plotly_convergence(df: pd.DataFrame, aggregate: bool = False) -> go.Figure:
    fig = go.Figure()
    if aggregate:
        for (m, p), g in df.groupby(["method_name", "problem_name"]):
            summary = (
                g.groupby("eval")["cummax_fitness"].agg(["mean", "std"]).reset_index()
            )
            fig.add_trace(
                go.Scatter(
                    x=summary["eval"],
                    y=summary["mean"],
                    mode="lines",
                    name=f"{m}-{p}",
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=summary["eval"],
                    y=summary["mean"] + summary["std"],
                    mode="lines",
                    line=dict(width=0),
                    showlegend=False,
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=summary["eval"],
                    y=summary["mean"] - summary["std"],
                    mode="lines",
                    line=dict(width=0),
                    fill="tonexty",
                    showlegend=False,
                )
            )
    else:
        for (m, p, s), g in df.groupby(["method_name", "problem_name", "seed"]):
            fig.add_trace(
                go.Scatter(
                    x=g["eval"],
                    y=g["cummax_fitness"],
                    mode="lines",
                    name=f"{m}-{p}-seed{s}",
                )
            )
    fig.update_layout(xaxis_title="Evaluations", yaxis_title="Best fitness")
    return fig


RESULTS_DIR = "results"


def discover_experiments(root=RESULTS_DIR):
    """Return list of experiment directories containing an experimentlog.jsonl."""
    exps = []
    if os.path.isdir(root):
        for entry in os.listdir(root):
            path = os.path.join(root, entry)
            if os.path.isdir(path) and os.path.exists(
                os.path.join(path, "progress.json")
            ):
                exps.append(entry)
    return sorted(exps)


def read_progress(exp_dir):
    path = os.path.join(exp_dir, "progress.json")
    if os.path.exists(path):
        with open(path) as f:
            return json.load(f)
    return None


def run() -> None:
    st.set_page_config(
        page_title="BLADE Experiment Browser",
        layout="wide",
        menu_items={"Get Help": "https://xai-liacs.github.io/BLADE/"},
    )

    if "last_refresh" not in st.session_state:
        st.session_state["last_refresh"] = time.time()
    elif time.time() - st.session_state["last_refresh"] > 60:
        st.session_state["last_refresh"] = time.time()
        st.rerun()

    logo_html = f"""
    <picture>
      <source media='(prefers-color-scheme: dark)' srcset='{LOGO_DARK}'>
      <source media='(prefers-color-scheme: light)' srcset='{LOGO_LIGHT}'>
      <img src='{LOGO_LIGHT}' style='width: 150px;'>
    </picture>
    """
    st.sidebar.markdown(logo_html, unsafe_allow_html=True)
    st.sidebar.markdown("### Experiments")

    experiments = discover_experiments()
    selected = st.session_state.get("selected_exp")

    if experiments:
        for exp in experiments:
            prog = read_progress(os.path.join(RESULTS_DIR, exp))
            icon = "✅" if prog and prog.get("end_time") else "⏳"
            if st.sidebar.button(exp, key=f"btn_{exp}", type="tertiary", icon=icon):
                st.session_state["selected_exp"] = exp
                selected = exp
            if prog and not prog.get("end_time"):
                total = prog.get("total", 1)
                pct = prog.get("current", 0) / total if total else 0
                st.sidebar.progress(pct)
    else:
        st.sidebar.write("No experiments found in 'results'.")

    if selected is None and experiments:
        selected = experiments[0]

    if selected:
        exp_dir = os.path.join(RESULTS_DIR, selected)
        logger = ExperimentLogger(exp_dir, read=True)

        prog = read_progress(exp_dir)
        if prog:
            if prog.get("end_time"):
                st.success("Finished")
            else:
                total = prog.get("total", 1)
                pct = prog.get("current", 0) / total if total else 0
                st.progress(pct)
            if prog.get("runs"):
                st.markdown("#### Run Progress")
                for r in prog["runs"]:
                    label = (
                        f"{r['method_name']} | {r['problem_name']} | seed {r['seed']}"
                    )
                    budget = r.get("budget", 1)
                    evaluations = r.get("evaluations", 0)
                    pct_run = evaluations / budget if budget else 0
                    st.progress(pct_run, text=label)

        st.header(f"Convergence - {selected}")
        df = convergence_dataframe(logger)
        if not df.empty:
            methods = sorted(df["method_name"].unique())
            problems = sorted(df["problem_name"].unique())
            method_sel = st.multiselect("Methods", methods, default=methods)
            problem_sel = st.multiselect("Problems", problems, default=problems)
            aggregate = st.checkbox("Aggregate runs", value=True)
            df_filt = df[
                df["method_name"].isin(method_sel)
                & df["problem_name"].isin(problem_sel)
            ]
            fig = plotly_convergence(df_filt, aggregate=aggregate)
            st.plotly_chart(fig, use_container_width=True)

            final_df = (
                df_filt.groupby(["method_name", "problem_name", "seed"])
                .last()
                .reset_index()
            )
            box_fig = go.Figure()
            for prob in problem_sel:
                subset = final_df[final_df["problem_name"] == prob]
                box_fig.add_trace(
                    go.Box(
                        y=subset["fitness"],
                        x=subset["method_name"],
                        name=prob,
                    )
                )
            box_fig.update_layout(yaxis_title="Fitness", xaxis_title="Method")
            st.plotly_chart(box_fig, use_container_width=True)

            st.markdown("#### Top Solutions")
            runs = logger.get_data()
            for m in method_sel:
                m_runs = runs[runs["method_name"] == m]
                m_runs = m_runs.sort_values(
                    by=["solution"],
                    key=lambda col: col.apply(
                        lambda s: s.get("fitness", -float("inf"))
                    ),
                    ascending=False,
                )
                top = m_runs.head(3)
                with st.expander(m):
                    for _, row in top.iterrows():
                        sol = row["solution"]
                        fname = f"{sol.get('name','sol')}.py"
                        st.write(f"{sol.get('name')} - {sol.get('fitness')}")
                        st.download_button(
                            label=f"Download {fname}",
                            data=sol.get("code", ""),
                            file_name=fname,
                            mime="text/x-python",
                            key=f"download_{m}_{sol.get('id', 'id')}",
                        )
        else:
            st.write("No data")
    else:
        st.write("Select an experiment from the sidebar to view results.")


def main() -> None:
    subprocess.run(["streamlit", "run", str(Path(__file__))], check=True)


if __name__ == "__main__":
    run()
